from __future__ import annotations

import hashlib
import json
import logging
import re
from collections import namedtuple, defaultdict
from functools import reduce
from typing import Optional, Generator, Any, Dict, List, Set, Tuple, Union

from attrs import define
from networkx import DiGraph, MultiDiGraph, all_shortest_paths, is_directed_acyclic_graph

from resotocore.ids import NodeId
from resotocore.model.model import (
    Model,
    Kind,
    AnyKind,
    ComplexKind,
    ArrayKind,
    DateTimeKind,
    DictionaryKind,
    StringKind,
    Property,
    SimpleKind,
    DateKind,
    DurationKind,
    UsageDatapoint,
    UsageMetricValues,
)
from resotocore.model.resolve_in_graph import GraphResolver, NodePath, ResolveProp
from resotocore.model.typed_model import from_js
from resotocore.types import Json, EdgeType, JsonElement
from resotocore.util import utc, utc_str, value_in_path, set_value_in_path, value_in_path_get

log = logging.getLogger(__name__)

# This version is used when the content hash of a node is computed.
# All computed hashes will be invalidated, by incrementing the version.
# This can be used, if computed values should be recomputed for all imported data.
ContentHashVersion = 3


class Section:
    # The reported section contains the data gathered by the collector.
    # This data is usually not changed by the user directly, but implicitly via changes on the
    # infrastructure, so the next collect run will change this state.
    reported = "reported"

    # This section holds changes that should be reflected by the given node.
    # The desired section can be queried the same way as the reported section
    # and allows to query commands of the graph with a common desired state.
    # For example the clean flag is manifested in the desired section.
    # The separate clean step would query all nodes that should be cleaned
    # and can compute the correct order of action by walking the graph structure.
    desired = "desired"

    # This section holds information about this node that are gathered during the import process.
    # Example: This section resolves common graph attributes like cloud, account, region, zone to make
    # querying the graph easy.
    metadata = "metadata"

    # Following sections are used to lookup special kinds in the graph hierarchy to simplify access.
    # See GraphResolver for details.
    # All resolved ancestors are written to this section.
    ancestors = "ancestors"

    # Resolved descendants would be written to this section.
    # Only here for completeness - currently not used.
    descendants = "descendants"

    # Usage of the resource
    usage = "usage"

    # The set of all content sections
    content_ordered = [reported, desired, metadata]
    content = set(content_ordered)

    # The list of all lookup sections
    lookup_sections_ordered = [ancestors, descendants, usage]

    # The list of all sections
    all_ordered = [*content_ordered, *lookup_sections_ordered]
    all = set(all_ordered)

    # remove the section plus dot if it exists in the string: reported.foo => foo
    __no_section = re.compile("^/?(" + "|".join(f"({s})" for s in content_ordered) + ")[.]")

    @classmethod
    def without_section(cls, path: str) -> str:
        return cls.__no_section.sub("", path, 1)


class EdgeTypes:
    # This edge type defines the default relationship between resources.
    # It is the main edge type and is assumed, if no edge type is given.
    # The related graph is also used as source of truth for graph updates.
    default: EdgeType = "default"

    # This edge type defines the order of delete operations.
    # A resource can be deleted, if all outgoing resources are deleted.
    delete: EdgeType = "delete"

    # The set of all allowed edge types.
    # Note: the database schema has to be adapted to support additional edge types.
    all: Set[EdgeType] = {default, delete}


class Direction:
    # Opposite direction as the edge direction.
    inbound = "in"
    # Same direction as the edge direction
    outbound = "out"
    # Ignore the direction of the edge and traverse in any direction.
    any = "any"

    # The set of all allowed directions.
    all: Set[str] = {inbound, outbound, any}


EdgeKey = namedtuple("EdgeKey", ["from_node", "to_node", "edge_type"])


@define
class BySearchCriteria:
    query: str


@define
class ByNodeId:
    value: NodeId


NodeSelector = Union[ByNodeId, BySearchCriteria]


@define
class DeferredEdge:
    from_node: NodeSelector
    to_node: NodeSelector
    edge_type: str


class GraphBuilder:
    def __init__(self, model: Model, change_id: str):
        self.model = model
        self.graph = MultiDiGraph()
        self.nodes = 0
        self.edges = 0
        self.deferred_edges: List[DeferredEdge] = []
        self.usage: List[UsageDatapoint] = []
        self.at = int(utc().timestamp())
        self.change_id = change_id

    def add_from_json(self, js: Json) -> None:
        if "id" in js and Section.reported in js:
            usage_json = js.get(Section.usage, {})
            if len(usage_json) == 0:
                usage_json = None
            self.add_node(
                node_id=js["id"],
                reported=js[Section.reported],
                desired=js.get(Section.desired, None),
                metadata=js.get(Section.metadata, None),
                search=js.get("search", None),
                replace=js.get("replace", False) is True,
            )
            if usage_json:
                values = {
                    k: UsageMetricValues(
                        min=v.get("min", v["avg"]),
                        avg=v["avg"],
                        max=v.get("max", v["avg"]),
                    )
                    for k, v in usage_json.items()
                }
                usage = UsageDatapoint(
                    id=js["id"],
                    change_id=self.change_id,
                    at=self.at,
                    v=values,
                )
                self.usage.append(usage)
        elif "from" in js and "to" in js:
            self.add_edge(js["from"], js["to"], js.get("edge_type", EdgeTypes.default))
        elif "from_selector" in js and "to_selector" in js:

            def parse_selector(js: Json) -> NodeSelector:
                if "node_id" in js:
                    return ByNodeId(NodeId(from_js(js["node_id"], str)))
                elif "search_criteria" in js:
                    return BySearchCriteria(from_js(js["search_criteria"], str))
                else:
                    raise AttributeError(f"can't parse edge selector! Got {json.dumps(js)}")

            self.store_deferred_connection(
                parse_selector(js["from_selector"]),
                parse_selector(js["to_selector"]),
                js.get("edge_type", EdgeTypes.default),
            )
        else:
            raise AttributeError(f"Format not understood! Got {json.dumps(js)} which is neither vertex nor edge.")

    def __update_property_size(self, kind: Kind, element: JsonElement) -> None:
        def prop_size(prop: Property, pk: Kind, part: JsonElement) -> None:
            if part is None:
                pass
            elif isinstance(pk, (StringKind, DateTimeKind, DateKind, DurationKind)) and isinstance(part, str):
                str_len = len(part)
                if prop.metadata is None:
                    prop.metadata = {}
                size = prop.metadata.get("len", 0)
                prop.metadata["len"] = max(size, str_len)
            elif isinstance(pk, SimpleKind):
                pass
            elif isinstance(pk, ArrayKind) and isinstance(part, list) and isinstance(pk.inner, StringKind):
                for elem in part:
                    prop_size(prop, pk.inner, elem)
            elif isinstance(pk, ArrayKind) and isinstance(part, list):
                for elem in part:
                    self.__update_property_size(pk.inner, elem)
            elif isinstance(pk, DictionaryKind) and isinstance(part, dict) and isinstance(pk.value_kind, StringKind):
                for elem in part.values():
                    prop_size(prop, pk.value_kind, elem)
            elif isinstance(pk, DictionaryKind) and isinstance(part, dict):
                for k, v in part.items():
                    self.__update_property_size(pk.key_kind, k)
                    self.__update_property_size(pk.value_kind, v)
            elif isinstance(pk, ComplexKind) and isinstance(part, dict):
                for cp, cpk in pk.all_props_with_kind():
                    prop_size(cp, cpk, part.get(cp.name, None))

        if isinstance(kind, ComplexKind) and isinstance(element, dict):
            prop_size(Property("root", kind.fqn), kind, element)

    def add_node(
        self,
        node_id: NodeId,
        reported: Json,
        desired: Optional[Json] = None,
        metadata: Optional[Json] = None,
        search: Optional[str] = None,
        replace: bool = False,
    ) -> None:
        self.nodes += 1
        # validate kind of this reported json
        coerced = self.model.check_valid(reported)
        reported = reported if coerced is None else coerced
        kind = self.model[reported]
        # create content hash
        sha = GraphBuilder.content_hash(reported, desired, metadata)
        # flat all properties into a single string for search
        flat = search if isinstance(search, str) else (GraphBuilder.flatten(reported, kind))
        self.graph.add_node(
            node_id,
            id=node_id,
            reported=reported,
            desired=desired,
            metadata=metadata,
            hash=sha,
            kind=kind,
            kinds=list(kind.kind_hierarchy()),
            kinds_set=kind.kind_hierarchy(),
            flat=flat,
            replace=replace | metadata.get("replace", False) is True if metadata else False,
        )
        # update property sizes
        self.__update_property_size(kind, reported)

    def add_edge(self, from_node: str, to_node: str, edge_type: EdgeType) -> None:
        self.edges += 1
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        self.graph.add_edge(from_node, to_node, key, edge_type=edge_type)

    def store_deferred_connection(self, from_selector: NodeSelector, to_selector: NodeSelector, edge_type: str) -> None:
        self.deferred_edges.append(DeferredEdge(from_selector, to_selector, edge_type))

    @staticmethod
    def content_hash(js: Json, desired: Optional[Json] = None, metadata: Optional[Json] = None) -> str:
        sha256 = hashlib.sha256()
        # all content hashes will be different, when the version changes
        sha256.update(ContentHashVersion.to_bytes(2, "big"))
        sha256.update(json.dumps(js, sort_keys=True).encode("utf-8"))
        if desired:
            sha256.update(json.dumps(desired, sort_keys=True).encode("utf-8"))
        if metadata:
            sha256.update(json.dumps(metadata, sort_keys=True).encode("utf-8"))
        return sha256.hexdigest()

    @staticmethod
    def flatten(js: Json, kind: Kind) -> str:
        result = ""

        def dispatch(value: Any, k: Kind) -> None:
            nonlocal result
            if isinstance(value, dict):
                for prop, elem in value.items():
                    sub = (
                        k.property_kind_of(prop, AnyKind())
                        if isinstance(k, ComplexKind)
                        else (k.value_kind if isinstance(k, DictionaryKind) else AnyKind())
                    )
                    dispatch(elem, sub)
            elif isinstance(value, list):
                sub = k.inner if isinstance(k, ArrayKind) else AnyKind()
                for elem in value:
                    dispatch(elem, sub)
            elif value is None or isinstance(value, bool):
                pass
            else:
                # in case of date time: "2017-05-30T22:04:34Z" -> "2017-05-30 22:04:34"
                if isinstance(k, DateTimeKind):
                    value = re.sub("[ZT]", " ", value)
                if result:
                    result += " "
                result += str(value).strip()

        dispatch(js, kind)
        return result

    def check_complete(self) -> None:
        # check that all vertices are given, that were defined in any edge definition
        # note: DiGraph will create an empty vertex node automatically
        for node_id, node in self.graph.nodes(data=True):
            assert node.get(Section.reported), f"{node_id} was used in an edge definition but not provided as vertex!"

        edge_types: Set[str] = {edge[2] for edge in self.graph.edges(data="edge_type")}
        al = EdgeTypes.all
        assert not edge_types.difference(al), f"Graph contains unknown edge types! Given: {edge_types}. Known: {al}"
        # make sure there is only one root node
        rid = GraphAccess.root_id(self.graph)
        root_node = self.graph.nodes[rid]

        # make sure the root
        if value_in_path(root_node, NodePath.reported_kind) == "graph_root" and rid != "root":
            # remove node with wrong id +
            root_node = self.graph.nodes[rid]
            root_node["id"] = "root"
            self.graph.add_node("root", **root_node)

            for succ in list(self.graph.successors(rid)):
                for edge_type in EdgeTypes.all:
                    key = GraphAccess.edge_key(rid, succ, edge_type)
                    if self.graph.has_edge(rid, succ, key):
                        self.graph.remove_edge(rid, succ, key)
                        self.add_edge("root", succ, edge_type)
            self.graph.remove_node(rid)


NodeData = Tuple[str, Json, Optional[Json], Optional[Json], Optional[Json], str, List[str], str]


class GraphAccess:
    def __init__(
        self,
        sub: MultiDiGraph,
        maybe_root_id: Optional[str] = None,
        visited_nodes: Optional[Set[NodeId]] = None,
        visited_edges: Optional[Set[EdgeKey]] = None,
    ):
        super().__init__()
        self.g = sub
        self.nodes = sub.nodes()
        self.visited_nodes: Set[NodeId] = visited_nodes if visited_nodes else set()
        self.visited_edges: Set[EdgeKey] = visited_edges if visited_edges else set()
        self.at = utc()
        self.at_json = utc_str(self.at)
        self.maybe_root_id = maybe_root_id
        self.resolved = False

    def root(self) -> str:
        return self.maybe_root_id if self.maybe_root_id else GraphAccess.root_id(self.g)

    def node(self, node_id: NodeId) -> Optional[Json]:
        self.visited_nodes.add(node_id)
        if self.g.has_node(node_id):
            n = self.nodes[node_id]
            return self.dump(node_id, n)
        else:
            return None

    def has_edge(self, from_id: object, to_id: object, edge_type: EdgeType) -> bool:
        key = self.edge_key(from_id, to_id, edge_type)
        result: bool = self.g.has_edge(from_id, to_id, key)
        if result:
            self.visited_edges.add(key)
        return result

    def resolve(self) -> None:
        if not self.resolved:
            self.resolved = True
            log.info("Resolve attributes in graph")
            for node_id in self.nodes:
                self.__resolve(node_id, self.nodes[node_id])
            self.__resolve_count_descendants()
            log.info("Resolve attributes finished.")

    def __resolve_count_descendants(self) -> None:
        visited: Set[str] = set()

        def count_successors_by(node_id: NodeId, edge_type: EdgeType, path: List[str]) -> Dict[str, int]:
            result: Dict[str, int] = {}
            to_visit = list(self.successors(node_id, edge_type))
            while to_visit:
                visit_next: List[NodeId] = []
                for elem_id in to_visit:
                    if elem_id not in visited:
                        visited.add(elem_id)
                        elem = self.nodes[elem_id]
                        if not value_in_path_get(elem, NodePath.is_phantom, False):
                            extracted = value_in_path(elem, path)
                            if isinstance(extracted, str):
                                result[extracted] = result.get(extracted, 0) + 1
                        # check if there is already a successor summary: stop the traversal and take the result.
                        existing = value_in_path(elem, NodePath.descendant_summary)
                        if existing and isinstance(existing, dict):
                            for summary_item, count in existing.items():
                                result[summary_item] = result.get(summary_item, 0) + count
                        else:
                            visit_next.extend(a for a in self.successors(elem_id, edge_type) if a not in visited)
                to_visit = visit_next
            return result

        for on_kind, prop in GraphResolver.count_successors.items():
            for node_id, node in self.g.nodes(data=True):
                kinds = node.get("kinds_set")
                if kinds and on_kind in kinds:
                    summary = count_successors_by(node_id, EdgeTypes.default, prop.extract_path)
                    set_value_in_path(summary, prop.to_path, node)
                    total = reduce(lambda left, right: left + right, summary.values(), 0)
                    set_value_in_path(total, NodePath.descendant_count, node)

    def __resolve(self, node_id: NodeId, node: Json) -> Json:
        def with_ancestor(ancestor: Json, prop: ResolveProp) -> None:
            extracted = value_in_path(ancestor, prop.extract_path)
            if extracted:
                set_value_in_path(extracted, prop.to_path, node)

        for resolver in GraphResolver.to_resolve:
            # search for ancestor that matches filter criteria
            anc = self.ancestor_of(node_id, EdgeTypes.default, resolver.kind)
            if anc:
                on_self = anc.get("id") == node_id
                for res in resolver.resolve:
                    if not on_self or res.apply_on_self:
                        with_ancestor(anc, res)
        return node

    def dump(self, node_id: NodeId, node: Json) -> Json:
        kind = node.get("kind", AnyKind())
        return self.dump_direct(node_id, node, kind)

    def predecessors(self, node_id: NodeId, edge_type: EdgeType) -> Generator[NodeId, Any, None]:
        for pred_id in self.g.predecessors(node_id):
            # direction from parent node to provided node
            if self.g.has_edge(pred_id, node_id, self.edge_key(pred_id, node_id, edge_type)):
                yield pred_id

    def successors(self, node_id: NodeId, edge_type: EdgeType) -> Generator[NodeId, Any, None]:
        for succ_id in self.g.successors(node_id):
            # direction from provided node to successor node
            if self.g.has_edge(node_id, succ_id, self.edge_key(node_id, succ_id, edge_type)):
                yield succ_id

    def ancestor_of(self, node_id: NodeId, edge_type: EdgeType, kind: str) -> Optional[Json]:
        # note: we are using breadth first search here on purpose.
        # if there is an ancestor with less distance to this node, we should use this one
        next_level = [node_id]

        while next_level:
            parents: List[NodeId] = []
            for p_id in next_level:
                p: Json = self.nodes[p_id]
                kinds: Optional[List[str]] = value_in_path(p, NodePath.kinds)
                if kinds and kind in kinds:
                    return p
                else:
                    parents.extend(self.predecessors(p_id, edge_type))
            next_level = parents
        return None

    def is_acyclic_per_edge_type(self) -> bool:
        """
        Checks if the graph is acyclic with respect to a specific edge type.
        This means it is valid if there are cycles in the graph but not for the same edge type.
        :return: True if the graph is acyclic for all edge types, otherwise False.
        """
        edges_per_type = defaultdict(list)
        # edge is a tuple: (from_node, to_node, edge_key)
        for edge in self.g.edges(keys=True):
            key: EdgeKey = edge[2]
            edges_per_type[key.edge_type].append(edge)
        for edges in edges_per_type.values():
            typed_graph = self.g.edge_subgraph(edges)
            acyclic = is_directed_acyclic_graph(typed_graph)
            if not acyclic:
                return False
        return True

    @staticmethod
    def dump_direct(node_id: NodeId, node: Json, kind: Kind, recompute: bool = False) -> Json:
        reported = node[Section.reported]
        desired: Optional[Json] = node.get(Section.desired, None)
        metadata: Optional[Json] = node.get(Section.metadata, None)
        if "id" not in node:
            node["id"] = node_id
        if recompute or "hash" not in node:
            node["hash"] = GraphBuilder.content_hash(reported, desired, metadata)
        if recompute or "flat" not in node:
            node["flat"] = GraphBuilder.flatten(reported, kind)
        if "kinds" not in node:
            node["kinds"] = [reported["kind"]]
        return node

    def not_visited_nodes(self) -> Generator[Json, None, None]:
        return (self.dump(nid, self.nodes[nid]) for nid in self.g.nodes if nid not in self.visited_nodes)

    def not_visited_edges(self, edge_type: EdgeType) -> Generator[Tuple[str, str], None, None]:
        # edge collection with (from, to, type): filter and drop type -> (from, to)
        edges = self.g.edges(data="edge_type")
        return (edge[:2] for edge in edges if edge[2] == edge_type and edge not in self.visited_edges)

    @staticmethod
    def edge_key(from_node: object, to_node: object, edge_type: EdgeType) -> EdgeKey:
        return EdgeKey(from_node, to_node, edge_type)

    @staticmethod
    def root_id(graph: DiGraph) -> NodeId:
        # noinspection PyTypeChecker
        roots: List[NodeId] = [n for n, d in graph.in_degree if d == 0]
        assert len(roots) == 1, f"Given subgraph has more than one root: {roots}"
        return roots[0]

    @staticmethod
    def merge_graphs(
        graph: MultiDiGraph,
    ) -> Tuple[List[str], GraphAccess, Generator[Tuple[str, GraphAccess], None, None]]:
        """
        Find all merge graphs in the provided graph.
        A merge graph is a self contained graph under a node which is marked with replace=true.
        Such nodes are replaced with the replace node in the database.
        Example:
        A -> B -> C(replace=true) -> E -> E1 -> E2
                                  -> F -> F1
               -> D(replace=true) -> G -> G1 -> G2 -> G3 -> G4

        This will result in 2 merge roots:
            C: [A, B]
            D: [A, B]

        Note that all successors of a merge node that is also a predecessors of the merge node is sorted out.

        :param graph: the incoming multi graph update.
        :return: the list of all merge roots, the expected parent graph and all merge root graphs.
        """

        # Find replace nodes: all nodes that are marked as replace node.
        # This method returns all replace roots as key, with the respective predecessors nodes as value.
        def replace_roots() -> Dict[NodeId, Set[NodeId]]:
            graph_root = GraphAccess.root_id(graph)
            replace_nodes: Dict[NodeId, Json] = {
                node_id: data for node_id, data in graph.nodes(data=True) if data.get("replace", False)
            }
            assert (
                len(replace_nodes) > 0
            ), "No replace nodes provided in the graph. Mark at least one node with replace=true!"
            result: Dict[NodeId, Set[NodeId]] = {}
            for node, data in replace_nodes.items():
                kind = GraphResolver.resolved_kind(data)
                assert (
                    kind is not None
                ), f"Node {node} is marked as replace node, but the kind is not resolved during import!"
                # compute the shortest path from root to here
                pres: Set[NodeId] = reduce(
                    lambda res, p: {*res, *p}, all_shortest_paths(graph, graph_root, node), set[NodeId]()
                )
                result[node] = pres
            # make sure there is no replace node beyond another replace node
            rs = result.copy()
            for node in rs:
                for nid, parent_nodes in rs.items():
                    if nid != node and node in parent_nodes:
                        log.info(f"Node {nid} marked as replace, but is child of another replace node {node}. Ignore.")
                        result.pop(nid, None)
            return result

        # Walk the graph from given starting node and return all successors.
        # A successor which is also a predecessors is not followed.
        def sub_graph_nodes(from_node: NodeId, parent_ids: Set[NodeId]) -> Set[NodeId]:
            to_visit = [from_node]
            visited: Set[NodeId] = {from_node}

            def successors(node: NodeId) -> List[NodeId]:
                return [a for a in graph.successors(node) if a not in visited and a not in parent_ids]

            while to_visit:
                to_visit = reduce(lambda li, node: li + successors(node), to_visit, list[NodeId]())
                visited.update(to_visit)
            return visited

        # Create a generator for all given merge roots by:
        #   - creating the set of all successors
        #   - creating a subgraph which contains all predecessors and all succors
        #   - all predecessors are marked as visited
        #   - all predecessors edges are marked as visited
        # This way it is possible to have nodes in the graph that will not be touched by the update
        # while edges will be created from successors of the merge node to predecessors of the merge node.
        def merge_sub_graphs(
            root_nodes: Dict[NodeId, Set[NodeId]], parent_nodes: Set[NodeId], parent_edges: Set[EdgeKey]
        ) -> Generator[Tuple[NodeId, GraphAccess], None, None]:
            all_successors: Set[NodeId] = set()
            for root, predecessors in root_nodes.items():
                successors: Set[NodeId] = sub_graph_nodes(root, predecessors)
                # make sure nodes are not "mixed" between different merge nodes
                overlap = successors & all_successors
                if overlap:
                    raise AttributeError(f"Nodes are referenced in more than one merge node: {overlap}")
                all_successors.update(successors)
                # create subgraph with all successors and all parents, where all parents are already marked as visited
                sub = GraphAccess(graph.subgraph(successors), root, parent_nodes, parent_edges)
                yield root, sub

        GraphAccess(graph).resolve()  # resolve graph references
        roots = replace_roots()
        parents: Set[NodeId] = reduce(lambda res, ps: {*res, *ps}, roots.values(), set[NodeId]())
        parent_graph = graph.subgraph(parents)
        graphs = merge_sub_graphs(roots, parents, set(parent_graph.edges(data="edge_type")))
        return list(roots.keys()), GraphAccess(parent_graph, GraphAccess.root_id(graph)), graphs



import asyncio
from datetime import timedelta

import pytest

from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.task.start_workflow_on_first_subscriber import wait_and_start
from resotocore.task.task_description import Workflow
from tests.resotocore.task.job_handler_test import InMemJobHandler


@pytest.mark.asyncio
async def test_wait_and_start(message_bus: MessageBus, test_workflow: Workflow) -> None:
    handler = InMemJobHandler()
    task = wait_and_start([test_workflow], handler, message_bus, timedelta(seconds=0))
    await asyncio.sleep(0.1)
    await message_bus.emit_event(CoreMessage.Connected, {"subscriber_id": "test", "channels": ["collect"]})
    await task
    assert handler.started_tasks == ["test_workflow"]



import pytest

from resotocore.analytics.posthog import PostHogEventSender
from resotocore.db import SystemData
from resotocore.util import utc


@pytest.mark.asyncio
async def test_send_analytics_proper() -> None:
    sd = SystemData("test", utc(), 1)
    async with PostHogEventSender(sd, client_flush_interval=0.01, client_retries=0) as sender:
        event = await sender.core_event("test-event")
        assert event.kind == "test-event"
    # reaching this point means: no exception has been thrown, which is the real test


@pytest.mark.asyncio
async def test_send_analytics_no_service() -> None:
    sd = SystemData("test", utc(), 1)
    async with PostHogEventSender(
        sd, flush_at=1, host="https://127.0.0.1:54321", client_flush_interval=0.01, client_retries=0
    ) as sender:
        event = await sender.core_event("test-event")
        assert event.kind == "test-event"
    # reaching this point means: no exception has been thrown, which is the real test



from __future__ import annotations

import json
import logging
import os
from os.path import basename
from typing import List, Optional, Dict, ClassVar

from attr import define, field

from resotocore.config import ConfigEntity
from resotocore.model.typed_model import from_js
from resotocore.report import (
    ReportSeverity,
    CheckConfigRoot,
    ReportCheck,
    Benchmark,
    Remediation,
    BenchmarkConfigRoot,
)
from resotocore.types import Json
from resotolib.core.model_export import dataclasses_to_resotocore_model

log = logging.getLogger(__name__)


@define
class ReportCheckConfig:
    kind: ClassVar[str] = "resoto_core_report_check"
    name: str = field(
        metadata={
            "description": "Name of this check. Must be unique within the provider and service.",
            "Example": "unused_elastic_ip",
        }
    )
    title: str = field(metadata={"description": "Title of the check."})
    result_kind: str = field(metadata={"description": "Resulting kind this check will emit. Example: aws_ec2_instance"})
    categories: List[str] = field(metadata={"description": "Categories of the check. Example: ['security', 'cost']"})
    severity: ReportSeverity = field(metadata={"description": "Severity of the check."})
    risk: str = field(metadata={"description": "What is the risk associated with related resources."})
    detect: Dict[str, str] = field(
        metadata={
            "description": "Defines possible detection methods.\n"
            "`resoto` defines a Resoto search, `resoto_cmd` a Resoto CLI command.\n "
            "At least one of `resoto` or `resoto_cmd` must be defined.\n"
            "Additional keys can be defined on top."
        }
    )
    remediation: Remediation = field(metadata={"description": "Remediation action for the check."})
    default_values: Optional[Json] = field(
        default=None,
        metadata={"description": "Default values for the check. Will be merged with the values from the config."},
    )
    url: Optional[str] = field(default=None, metadata={"description": "URL that documents the check."})
    related: Optional[List[str]] = field(default=None, metadata={"description": "List of related checks."})
    internal_notes: Optional[str] = field(default=None, metadata={"description": "Internal notes for the check."})


@define
class ReportCheckCollectionConfig:
    kind: ClassVar[str] = CheckConfigRoot
    provider: str = field(metadata={"description": "Cloud provider of all checks."})
    service: str = field(metadata={"description": "Cloud provider service of all checks."})
    checks: List[ReportCheckConfig] = field(factory=list, kw_only=True, metadata={"description": "List of checks."})

    @staticmethod
    def from_files() -> Dict[str, Json]:
        # load the checks from the report directory
        static_path = os.path.abspath(os.path.dirname(__file__) + "/../static/report/checks")
        result = {}
        if os.path.exists(static_path):
            for provider in (d.path for d in os.scandir(static_path) if d.is_dir()):
                for service in (d.path for d in os.scandir(provider) if d.is_file() and d.name.endswith(".json")):
                    with open(service, "rt", encoding="utf-8") as f:
                        result[basename(service).rsplit(".", maxsplit=1)[0]] = json.load(f)
        return result

    @staticmethod
    def from_config(cfg: ConfigEntity) -> List[ReportCheck]:
        return ReportCheckCollectionConfig.from_json(cfg.config[CheckConfigRoot])

    @staticmethod
    def from_json(js: Json) -> List[ReportCheck]:
        def report_check(pdr: str, svc: str, check: Json) -> ReportCheck:
            cr = check.copy()
            cr["provider"] = pdr
            cr["service"] = svc
            cr["id"] = f"{pdr}_{svc}_{check['name']}"
            return from_js(cr, ReportCheck)

        pdr = js["provider"]
        svc = js["service"]
        return [report_check(pdr, svc, check) for check in js["checks"]]


@define
class CheckCollectionConfig:
    kind: ClassVar[str] = "resoto_core_report_check_collection"
    title: str = field(metadata={"description": "Title of the benchmark or report check collection."})
    description: str = field(metadata={"description": "Description of the benchmark."})
    documentation: Optional[str] = field(default=None, kw_only=True, metadata={"description": "Documentation URL."})
    checks: Optional[List[str]] = field(
        default=None, kw_only=True, metadata={"description": "List of checks to perform."}
    )
    children: Optional[List[CheckCollectionConfig]] = field(
        default=None, kw_only=True, metadata={"description": "Nested collections."}
    )

    def is_valid(self, checks: Dict[str, ReportCheck]) -> bool:
        return (
            not (self.children and self.checks)
            and all(c in checks for c in self.checks or [])  # pylint: disable=not-an-iterable
            and all(c.is_valid(checks) for c in self.children or [])  # pylint: disable=not-an-iterable
        )


@define
class BenchmarkConfig(CheckCollectionConfig):
    kind: ClassVar[str] = BenchmarkConfigRoot

    id: str = field(metadata={"description": "Unique ID of the benchmark."})
    framework: str = field(metadata={"description": "Framework the benchmark is based on."})
    version: str = field(metadata={"description": "Version of the benchmark."})
    clouds: Optional[List[str]] = field(
        default=None,
        metadata={
            "description": "List of applicable cloud providers. "
            "If the benchmark is not cloud specific, the value would be null."
        },
    )

    @staticmethod
    def from_files() -> Dict[str, Json]:
        # load the benchmarks from the report directory
        static_path = os.path.abspath(os.path.dirname(__file__) + "/../static/report/benchmark")
        result = {}
        if os.path.exists(static_path):
            for provider in (d.path for d in os.scandir(static_path) if d.is_dir()):
                for path in (d.path for d in os.scandir(provider) if d.is_file() and d.name.endswith(".json")):
                    with open(path, "rt", encoding="utf-8") as f:
                        bid = basename(path).rsplit(".", maxsplit=1)[0]
                        benchmark = json.load(f)
                        benchmark["id"] = bid
                        result[bid] = benchmark
        return result

    @staticmethod
    def from_config(cfg: ConfigEntity) -> Benchmark:
        # Benchmark and BenchmarkConfig are structurally identical.
        # If Benchmark needs to change, the config is here to have a migration path.
        return from_js(cfg.config[BenchmarkConfigRoot], Benchmark)


def config_model() -> List[Json]:
    config_classes = {ReportCheckCollectionConfig, BenchmarkConfig}
    return dataclasses_to_resotocore_model(config_classes, use_optional_as_required=True)



from typing import cast, Awaitable, Dict
from resotocore.db.packagedb import PackageEntityDb, app_package_entity_db
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.infra_apps.manifest import AppManifest
from resotocore.ids import InfraAppName
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.config import ConfigHandler
from resotocore.cli.model import InfraAppAlias
from arango.database import StandardDatabase
from types import SimpleNamespace
import pytest
from asyncio import Future
import aiofiles
import aiohttp
from pathlib import Path


@pytest.fixture
async def package_entity_db(test_db: StandardDatabase) -> PackageEntityDb:
    async_db = AsyncArangoDB(test_db)
    entity_db = app_package_entity_db(async_db, "test_package_entity_db")
    await entity_db.create_update_schema()
    await entity_db.wipe()
    return entity_db


def async_none() -> Awaitable[None]:
    future: Future[None] = Future()
    future.set_result(None)
    return future


config_handler = cast(
    ConfigHandler,
    SimpleNamespace(
        get_config=lambda config_id: async_none(),
        put_config=lambda config_entity, validate: async_none(),
        delete_config=lambda config_id: async_none(),
        update_configs_model=lambda models: async_none(),
        put_config_validation=lambda config_validation: async_none(),
    ),
)


@pytest.mark.asyncio
async def test_install_delete(package_entity_db: PackageEntityDb) -> None:
    name = InfraAppName("cleanup-untagged")
    enabled_aliases: Dict[str, InfraAppAlias] = {}

    def enable_alias(alias: InfraAppAlias) -> None:
        enabled_aliases[alias.name] = alias

    def disable_alias(name: str) -> None:
        enabled_aliases.pop(name)

    package_manager = PackageManager(
        package_entity_db,
        config_handler,
        enable_alias,
        disable_alias,
    )
    await package_manager.start()

    manifest = await package_manager.install(name, None)
    assert manifest is not None
    assert manifest.name == name

    # check that it is installed
    installed_apps = [name async for name in package_manager.list()]
    assert installed_apps == [name]

    installed_app = await package_manager.get_manifest(name)
    assert installed_app is not None
    assert installed_app.name == name

    # check that the alias is created
    alias_template = enabled_aliases.get(name)
    assert alias_template is not None
    assert alias_template.name == manifest.name
    assert alias_template.template() == f"apps run {manifest.name}" + r" {{args}}"
    assert alias_template.description == manifest.description
    assert alias_template.readme == manifest.readme

    # update is possible
    updated_manifest = await package_manager.update(name)
    assert updated_manifest is not None
    assert updated_manifest.name == name

    # update all is possible
    updated_apps = [name async for name, manifest in package_manager.update_all() if isinstance(manifest, AppManifest)]
    installed_apps = [name async for name in package_manager.list()]
    assert installed_apps == updated_apps == [name]

    # check that it can be deleted
    await package_manager.delete(name)

    # check that it is not installed anymore
    installed_apps_after_deletion = [name async for name in package_manager.list()]
    assert installed_apps_after_deletion == []

    # check that the alias is deleted
    alias_template = enabled_aliases.get(name)
    assert alias_template is None


@pytest.mark.asyncio
async def test_local_install(package_entity_db: PackageEntityDb) -> None:
    name = InfraAppName("cleanup-untagged")
    package_manager = PackageManager(package_entity_db, config_handler, lambda at: None, lambda s: None)
    await package_manager.start()

    # check that the app is not installed
    installed_apps = [name async for name in package_manager.list()]
    assert installed_apps == []

    async with aiohttp.ClientSession() as session:
        async with session.get(package_manager.cdn_url) as response:
            assert response.status == 200
            index_bytes = await response.read()

            async with aiofiles.tempfile.TemporaryDirectory() as tmp:
                index_path = Path(tmp) / "index.json"

                async with aiofiles.open(index_path, "wb") as f:
                    await f.write(index_bytes)

                await package_manager.install(
                    name,
                    "file://" + str(index_path),
                )

    # check that the app is installed
    installed_apps = [name async for name in package_manager.list()]
    assert installed_apps == [name]



import sys

# All python version specific differences are listed here


# string.removesuffix is introduced in 3.9
if sys.version_info < (3, 9):

    def remove_suffix(s: str, suffix: str) -> str:
        if s.endswith(suffix):
            return s[: -len(suffix)]
        else:
            return s

else:

    def remove_suffix(s: str, suffix: str) -> str:
        return s.removesuffix(suffix)



import asyncio
from datetime import timedelta
from typing import Dict, List, Tuple

from pytest import mark

from resotocore.ids import TaskId
from resotocore.util import group_by
from resotocore.worker_task_queue import WorkerTaskDescription, WorkerTaskQueue, WorkerTask


@mark.asyncio
async def test_handle_work_successfully(
    task_queue: WorkerTaskQueue,
    worker: Tuple[WorkerTaskDescription, WorkerTaskDescription, WorkerTaskDescription],
    performed_by: Dict[str, List[str]],
) -> None:
    success_task, _, _ = worker

    all_tasks = [create_task(str(n), success_task.name) for n in range(0, 20)]
    for t in all_tasks:
        await task_queue.add_task(t)

    results = await asyncio.gather(*[a.callback for a in all_tasks])
    assert results == [{"result": "done!"} for _ in range(0, 20)]

    # make sure the work is split equally between all workers: 20 work items by 4 workers: 5 work items each
    by_worker = group_by(lambda x: x, (item for sublist in performed_by.values() for item in sublist))
    assert len(by_worker) == 4
    for work_done in by_worker.values():
        assert len(work_done) == 5


@mark.asyncio
async def test_handle_failure(
    task_queue: WorkerTaskQueue, worker: Tuple[WorkerTaskDescription, WorkerTaskDescription, WorkerTaskDescription]
) -> None:
    _, fail_task, _ = worker

    all_tasks = [create_task(str(n), fail_task.name) for n in range(0, 20)]
    for t in all_tasks:
        await task_queue.add_task(t)

    results = await asyncio.gather(*[a.callback for a in all_tasks], return_exceptions=True)
    # make sure all results are failures
    for r in results:
        assert isinstance(r, Exception)


@mark.asyncio
async def test_handle_outdated(
    task_queue: WorkerTaskQueue,
    worker: Tuple[WorkerTaskDescription, WorkerTaskDescription, WorkerTaskDescription],
    performed_by: Dict[str, List[str]],
) -> None:
    _, _, outdated_task = worker

    all_tasks = [create_task(str(n), outdated_task.name) for n in range(0, 20)]
    for t in all_tasks:
        await task_queue.add_task(t, retry_count=3)

    await asyncio.sleep(0)

    count_outstanding = 0
    while task_queue.outstanding_tasks:
        await task_queue.check_outdated_unassigned_tasks()
        count_outstanding += 1

    # every message is retried 3 times ==> 4 times to get rid of all messages
    assert count_outstanding == 4

    results = await asyncio.gather(*[a.callback for a in all_tasks], return_exceptions=True)
    # make sure all results are failures
    for r in results:
        assert isinstance(r, Exception)

    # 20 work items by 4 workers: 5 work items each + retried 3 times (15) => 20
    by_worker = group_by(lambda x: x, (item for sublist in performed_by.values() for item in sublist))
    assert len(by_worker) == 4
    for work_done in by_worker.values():
        assert len(work_done) == 20


def create_task(uid: str, name: str) -> WorkerTask:
    return WorkerTask(TaskId(uid), name, {}, {}, asyncio.get_event_loop().create_future(), timedelta())



import asyncio
import json
import logging
import os
import sqlite3
from datetime import timedelta
from functools import partial
from typing import List, Dict, Optional, Any, Tuple, Type, TypeVar, cast, Callable, Set

import pytest
import yaml
from _pytest.logging import LogCaptureFixture
from aiohttp import ClientTimeout
from aiohttp.web import Request
from aiostream import stream
from aiostream.core import Stream
from attrs import evolve
from pytest import fixture

from resotocore import version
from resotocore.cli import is_node
from resotocore.cli.cli import CLIService
from resotocore.cli.command import HttpCommand, JqCommand, AggregateCommand, all_commands
from resotocore.dependencies import Dependencies
from resotocore.cli.model import CLIContext, WorkerCustomCommand, CLI, FilePath
from resotocore.cli.tip_of_the_day import generic_tips
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem
from resotocore.db.graphdb import ArangoGraphDB
from resotocore.db.jobdb import JobDb
from resotocore.error import CLIParseError
from resotocore.graph_manager.graph_manager import GraphManager
from resotocore.ids import InfraAppName, GraphName
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.infra_apps.runtime import Runtime
from resotocore.model.model import Model
from resotocore.model.typed_model import to_js
from resotocore.query.model import Template, Query
from resotocore.report import Inspector
from resotocore.report.report_config import BenchmarkConfig
from resotocore.task.task_description import TimeTrigger, Workflow, EventTrigger
from resotocore.task.task_handler import TaskHandlerService
from resotocore.types import JsonElement, Json
from resotocore.user import UsersConfigId
from resotocore.util import AccessJson, utc_str, utc
from resotocore.worker_task_queue import WorkerTask
from tests.resotocore.util_test import not_in_path


@fixture
def json_source() -> str:
    nums = ",".join([f'{{ "num": {a}, "inner": {{"num": {a%10}}}}}' for a in range(0, 100)])
    return "json [" + nums + "," + nums + "]"


def test_known_category(dependencies: Dependencies) -> None:
    allowed_categories = {"search", "format", "action", "setup", "misc"}
    for cmd in all_commands(dependencies):
        assert cmd.category in allowed_categories, f"Unknown category {cmd.category} for command {cmd.name}"


@pytest.mark.asyncio
async def test_echo_source(cli: CLI) -> None:
    # no arg passed to json
    result = await cli.execute_cli_command("echo", stream.list)
    assert result[0] == [""]

    # simple string passed to json
    result = await cli.execute_cli_command("echo this is a string", stream.list)
    assert result[0] == ["this is a string"]

    result = await cli.execute_cli_command('echo   "foo bla bar" ', stream.list)
    assert result[0] == ["foo bla bar"]


@pytest.mark.asyncio
async def test_json_source(cli: CLI) -> None:
    # json object passed to json
    result = await cli.execute_cli_command('json {"a": 1}', stream.list)
    assert result[0] == [{"a": 1}]

    # json array passed to json
    result = await cli.execute_cli_command('json [{"a": 1}, {"b":2}]', stream.list)
    assert result[0] == [{"a": 1}, {"b": 2}]

    # json string passed to json
    result = await cli.execute_cli_command('json "foo bla bar"', stream.list)
    assert result[0] == ["foo bla bar"]


@pytest.mark.asyncio
async def test_predecessors(cli: CLI) -> None:
    r1 = await cli.execute_cli_command("search id(4_0) | predecessors", stream.list)
    assert len(r1[0]) == 1
    r2 = await cli.execute_cli_command("search id(4_0) | predecessors --with-origin", stream.list)
    assert len(r2[0]) == 2
    r3 = await cli.execute_cli_command("search id(4_0) | predecessors --with-origin default", stream.list)
    assert len(r3[0]) == 2
    r4 = await cli.execute_cli_command("search id(4_0) | predecessors delete", stream.list)
    assert len(r4[0]) == 0


@pytest.mark.asyncio
async def test_ancestors(cli: CLI) -> None:
    r1 = await cli.execute_cli_command("search id(4_0) | ancestors", stream.list)
    assert len(r1[0]) == 4
    r2 = await cli.execute_cli_command("search id(4_0) | ancestors --with-origin", stream.list)
    assert len(r2[0]) == 5
    r3 = await cli.execute_cli_command("search id(4_0) | ancestors --with-origin default", stream.list)
    assert len(r3[0]) == 5
    r4 = await cli.execute_cli_command("search id(4_0) | ancestors delete", stream.list)
    assert len(r4[0]) == 0


@pytest.mark.asyncio
async def test_successors(cli: CLI) -> None:
    r1 = await cli.execute_cli_command("search id(4) | successors", stream.list)
    assert len(r1[0]) == 10
    r2 = await cli.execute_cli_command("search id(4) | successors --with-origin", stream.list)
    assert len(r2[0]) == 11
    r3 = await cli.execute_cli_command("search id(4) | successors --with-origin default", stream.list)
    assert len(r3[0]) == 11
    r4 = await cli.execute_cli_command("search id(4) | successors delete", stream.list)
    assert len(r4[0]) == 0


@pytest.mark.asyncio
async def test_descendants(cli: CLI) -> None:
    r1 = await cli.execute_cli_command("search id(4) | descendants", stream.list)
    assert len(r1[0]) == 10
    r2 = await cli.execute_cli_command("search id(4) | descendants --with-origin", stream.list)
    assert len(r2[0]) == 11
    r3 = await cli.execute_cli_command("search id(4) | descendants --with-origin default", stream.list)
    assert len(r3[0]) == 11
    r4 = await cli.execute_cli_command("search id(4) | descendants delete", stream.list)
    assert len(r4[0]) == 0


@pytest.mark.asyncio
async def test_search_source(cli: CLIService) -> None:
    result = await cli.execute_cli_command('search is("foo") and some_int==0 --> identifier=~"9_"', stream.list)
    assert len(result[0]) == 10
    await cli.dependencies.template_expander.put_template(
        Template("test", 'is(foo) and some_int==0 --> identifier=~"{{fid}}"')
    )
    result2 = await cli.execute_cli_command('search expand(test, fid="9_")', stream.list)
    assert len(result2[0]) == 10

    result3 = await cli.execute_cli_command("search --with-edges is(graph_root) -[0:1]->", stream.list)
    # node: graph_root
    # node: collector
    # edge: graph_root -> collector
    # -----------------------------
    # = 3 elements
    assert len(result3[0]) == 3

    result4 = await cli.execute_cli_command("search --explain --with-edges is(graph_root) -[0:1]->", stream.list)
    assert result4[0][0]["rating"] == "simple"

    # use absolute path syntax
    result5 = await cli.execute_cli_command(
        "search aggregate(/reported.kind: sum(/reported.some_int) as si): "
        "is(foo) and not(/reported.some_int!=0) "
        "{child: --> /metadata!=null} some_int==0 "
        "with(any, --> /metadata!=null) sort /reported.name asc limit 1",
        stream.list,
    )
    assert result5 == [[{"group": {"kind": "foo"}, "si": 0}]]


@pytest.mark.asyncio
async def test_sleep_source(cli: CLI) -> None:
    with pytest.raises(CLIParseError):
        await cli.evaluate_cli_command("sleep forever")
    result = await cli.execute_cli_command("sleep 0.001; echo hello", stream.list)
    assert result == [[""], ["hello"]]


@pytest.mark.asyncio
async def test_count_command(cli: CLI, json_source: str) -> None:
    # count instances
    result = await cli.execute_cli_command(f"{json_source} | count", stream.list)
    assert len(result[0]) == 2
    assert result[0] == ["total matched: 200", "total unmatched: 0"]

    # count attributes
    result = await cli.execute_cli_command(f"{json_source} | count num", stream.list)
    assert len(result[0]) == 102
    assert result[0][-2] == "total matched: 200"
    assert result[0][-1] == "total unmatched: 0"

    # count attributes with path
    result = await cli.execute_cli_command(f"{json_source} | count inner.num", stream.list)
    assert len(result[0]) == 12
    assert result[0][-2] == "total matched: 200"
    assert result[0][-1] == "total unmatched: 0"

    # count unknown attributes
    result = await cli.execute_cli_command(f"{json_source} | count does_not_exist", stream.list)
    assert len(result[0]) == 2
    assert result[0] == ["total matched: 0", "total unmatched: 200"]


@pytest.mark.asyncio
async def test_head_command(cli: CLI) -> None:
    assert await cli.execute_cli_command("json [1,2,3,4,5] | head 2 | dump", stream.list) == [[1, 2]]
    assert await cli.execute_cli_command("json [1,2,3,4,5] | head -2 | dump", stream.list) == [[1, 2]]
    assert await cli.execute_cli_command("json [1,2,3,4,5] | head | dump", stream.list) == [[1, 2, 3, 4, 5]]


@pytest.mark.asyncio
async def test_tail_command(cli: CLI) -> None:
    assert await cli.execute_cli_command("json [1,2,3,4,5] | tail 2 | dump", stream.list) == [[4, 5]]
    assert await cli.execute_cli_command("json [1,2,3,4,5] | tail -2 | dump", stream.list) == [[4, 5]]
    assert await cli.execute_cli_command("json [1,2,3,4,5] | tail | dump", stream.list) == [[1, 2, 3, 4, 5]]


@pytest.mark.asyncio
async def test_chunk_command(cli: CLI, json_source: str) -> None:
    result: List[List[str]] = await cli.execute_cli_command(f"{json_source} | chunk 50 | dump", stream.list)
    assert len(result[0]) == 4  # 200 in chunks of 50
    for a in result[0]:
        assert len(a) == 50


@pytest.mark.asyncio
async def test_flatten_command(cli: CLI, json_source: str) -> None:
    result = await cli.execute_cli_command(f"{json_source} | chunk 50 | flatten", stream.list)
    assert len(result[0]) == 200


@pytest.mark.asyncio
async def test_uniq_command(cli: CLI, json_source: str) -> None:
    result = await cli.execute_cli_command(f"{json_source} | uniq", stream.list)
    assert len(result[0]) == 100


@pytest.mark.asyncio
async def test_set_desired_command(cli: CLI) -> None:
    result = await cli.execute_cli_command('search is("foo") | set_desired a="test" b=1 c=true | dump', stream.list)
    assert len(result[0]) == 11
    for elem in result[0]:
        assert {"a": "test", "b": 1, "c": True}.items() <= elem["desired"].items()


@pytest.mark.asyncio
async def test_set_metadata_command(cli: CLI) -> None:
    result = await cli.execute_cli_command('search is("foo") | set_metadata a="test" b=1 c=true | dump', stream.list)
    assert len(result[0]) == 11
    for elem in result[0]:
        assert {"a": "test", "b": 1, "c": True}.items() <= elem["metadata"].items()


@pytest.mark.asyncio
async def test_clean_command(cli: CLI) -> None:
    result = await cli.execute_cli_command('search is("foo") | clean | dump', stream.list)
    assert len(result[0]) == 11
    for elem in result[0]:
        assert {"clean": True}.items() <= elem["desired"].items()


@pytest.mark.asyncio
async def test_protect_command(cli: CLI) -> None:
    result = await cli.execute_cli_command('search is("foo") | protect | dump', stream.list)
    assert len(result[0]) == 11
    for elem in result[0]:
        assert {"protected": True}.items() <= elem["metadata"].items()


@pytest.mark.asyncio
async def test_list_sink(cli: CLI, dependencies: Dependencies) -> None:
    result = await cli.execute_cli_command("json [1,2,3] | dump", stream.list)
    assert result == [[1, 2, 3]]


@pytest.mark.asyncio
async def test_flat_sink(cli: CLI) -> None:
    parsed = await cli.evaluate_cli_command("json [1,2,3] | dump; json [4,5,6] | dump; json [7,8,9] | dump")
    result = await stream.list(stream.concat(stream.iterate((await p.execute())[1] for p in parsed)))
    assert result == [1, 2, 3, 4, 5, 6, 7, 8, 9]


@pytest.mark.asyncio
async def test_format(cli: CLI) -> None:
    # access properties by name and path
    result = await cli.execute_cli_command(
        'json {"a":"b", "b": {"c":"d"}} | format a:{a} b:{b.c} na:{fuerty}', stream.list
    )
    assert result[0] == ["a:b b:d na:null"]

    # use correct type
    props = dict(a="a", b=True, c=False, d=None, e=12, f=1.234)
    result = await cli.execute_cli_command(f"json {json.dumps(props)}" " | format {a}:{b}:{c}:{d}:{e}:{f}", stream.list)
    assert result[0] == ["a:true:false:null:12:1.234"]
    # access deeply nested properties with dict and array
    result = await cli.execute_cli_command(
        'json {"a":{"b":{"c":{"d":[0,1,2, {"e":"f"}]}}}} | format will be an >{a.b.c.d[3].e}<', stream.list
    )
    assert result[0] == ["will be an >f<"]
    # make sure any path that is not available leads to the null value
    result = await cli.execute_cli_command("json {} | format {a}:{b.c.d}:{foo.bla[23].test}", stream.list)
    assert result[0] == ["null:null:null"]

    # Queries that use the reported section, also interpret the format in the reported section
    result = await cli.execute_cli_command(
        "search id(sub_root) limit 1 | format {{aa}} {some_string} test}} {some_int} {/metadata.node_id} {{",
        stream.list,
    )
    assert result[0] == ["{aa} hello test} 0 sub_root {"]


@pytest.mark.asyncio
async def test_workflows_command(cli: CLIService, task_handler: TaskHandlerService, test_workflow: Workflow) -> None:
    async def execute(cmd: str) -> List[JsonElement]:
        ctx = CLIContext(cli.cli_env)
        return (await cli.execute_cli_command(cmd, stream.list, ctx))[0]  # type: ignore

    assert await execute("workflows list") == ["sleep_workflow", "wait_for_collect_done", "test_workflow"]
    assert await execute("workflows show test_workflow") == [to_js(test_workflow)]
    wf = await execute("workflows run test_workflow")
    assert wf[0].startswith("Workflow test_workflow started with id")  # type: ignore
    assert len(await execute("workflows running")) == 1

    # executing an already running workflow will give a specific message
    await execute("workflows run sleep_workflow")
    sf = await execute("workflows run sleep_workflow")
    assert sf[0].startswith("Workflow sleep_workflow already running with id ")  # type: ignore

    # make sure to wait for all tasks to finish
    for rt in await task_handler.running_tasks():
        await task_handler.delete_running_task(rt)

    # access the history of all workflows
    history = AccessJson.wrap_list(await execute("workflows history"))
    assert len(history) == 1
    assert history[0].sleep_workflow.count == 1
    assert history[0].test_workflow.count == 1

    # access the history of a specific workflow
    history_test = AccessJson.wrap_list(await execute("workflows history test_workflow"))
    assert len(history_test) == 1
    wf_run = history_test[0]
    assert all(n in wf_run for n in ["id", "task_started_at", "duration"])

    # access the log of a specific workflow run
    task_log = await execute(f"workflows log {wf_run['id']}")
    assert len(task_log) == 1


@pytest.mark.asyncio
async def test_jobs_command(cli: CLIService, task_handler: TaskHandlerService, job_db: JobDb) -> None:
    async def execute(cmd: str) -> List[List[JsonElement]]:
        ctx = CLIContext(cli.cli_env)
        return await cli.execute_cli_command(cmd, stream.list, ctx)

    # add job with schedule
    result = await execute('jobs add --id hello --schedule "23 1 * * *" echo Hello World @NOW@')
    assert result == [["Job hello added."]]
    job = await job_db.get("hello")
    assert job is not None
    assert job.command.command == "echo Hello World @NOW@"
    assert job.trigger == TimeTrigger("23 1 * * *")
    assert job.wait is None
    assert job in task_handler.task_descriptions
    assert job.environment == {"graph": "ns", "section": "reported"}

    # add job with schedule and event
    with_event = await execute('jobs add --id timed_hi --schedule "23 1 * * *" --wait-for-event foo echo Hello World')
    assert with_event == [["Job timed_hi added."]]
    job_with_event: Job = await job_db.get("timed_hi")  # type: ignore
    assert job_with_event.wait is not None
    event_trigger, timeout = job_with_event.wait
    assert event_trigger.message_type == "foo"
    assert timeout == timedelta(hours=1)
    assert job_with_event.environment == {"graph": "ns", "section": "reported"}
    assert job_with_event in task_handler.task_descriptions

    # add job with event
    only_event = await execute("jobs add --id only_event --wait-for-event foo echo Hello World")
    assert only_event == [["Job only_event added."]]
    job_only_event: Job = await job_db.get("only_event")  # type: ignore
    assert job_only_event.trigger == EventTrigger("foo")
    assert job_only_event.wait is None
    assert job_only_event.environment == {"graph": "ns", "section": "reported"}
    assert job_only_event in task_handler.task_descriptions

    # add job without any trigger
    no_trigger = await execute("jobs add --id no_trigger echo Hello World")
    assert no_trigger == [["Job no_trigger added."]]
    job_no_trigger: Job = await job_db.get("no_trigger")  # type: ignore
    assert job_no_trigger.wait is None
    assert job_no_trigger.environment == {"graph": "ns", "section": "reported"}
    assert job_no_trigger in task_handler.task_descriptions

    # deactivate timed_hi
    deactivated = await execute("jobs deactivate timed_hi")
    assert deactivated[0][0]["active"] is False  # type: ignore

    # activate timed_hi
    activated = await execute("jobs activate timed_hi")
    assert activated[0][0]["active"] is True  # type: ignore

    # show specific job
    no_trigger_show = await execute("jobs show no_trigger")
    assert len(no_trigger_show[0]) == 1

    # show all jobs
    all_jobs = await execute("jobs list")
    assert len(all_jobs[0]) == 4

    # start the job
    run_hello = await execute("jobs run timed_hi")
    assert run_hello[0][0].startswith("Job timed_hi started with id")  # type: ignore
    assert [t for t in await task_handler.running_tasks() if t.descriptor.id == "timed_hi"]

    # list all running jobs
    all_running = await execute("jobs running")
    assert [r["job"] for r in all_running[0]] == ["timed_hi"]  # type: ignore

    # delete a job
    deleted = await execute("jobs delete timed_hi")
    assert deleted == [["Job timed_hi deleted."]]


@pytest.mark.asyncio
async def test_tag_command(
    cli: CLIService, performed_by: Dict[str, List[str]], incoming_tasks: List[WorkerTask], caplog: LogCaptureFixture
) -> None:
    counter = 0

    def nr_of_performed() -> int:
        nonlocal counter
        performed = len(performed_by)
        increase = performed - counter
        counter = performed
        return increase

    nr_of_performed()  # reset to 0

    assert await cli.execute_cli_command("echo id_does_not_exist | tag update foo bla", stream.list) == [[]]
    assert nr_of_performed() == 0
    res1 = await cli.execute_cli_command(
        'json ["root", "collector"] | tag update foo "bla_{reported.some_int}" | dump', stream.list
    )
    assert nr_of_performed() == 2
    assert {a["id"] for a in res1[0]} == {"root", "collector"}
    assert len(incoming_tasks) == 2
    # check that the worker task data is correct
    data = AccessJson(incoming_tasks[0].data)
    assert data["update"] is not None  # tag update -> data.update is defined
    assert not data.node.reported.is_none  # the node reported section is defined
    assert not data.node.metadata.is_none  # the node metadata section is defined
    assert not data.node.ancestors.cloud.reported.is_none  # the ancestors cloud section is defineda
    assert data["update"].foo == "bla_0"  # using the renderer bla_{reported.some_int}
    res2 = await cli.execute_cli_command('search is("foo") | tag update foo bla', stream.list)
    assert nr_of_performed() == 11
    assert len(res2[0]) == 11
    res2_tag_no_val = await cli.execute_cli_command('search is("foo") | tag update foobar', stream.list)
    assert nr_of_performed() == 11
    assert len(res2_tag_no_val[0]) == 11
    res3 = await cli.execute_cli_command('search is("foo") | tag delete foo', stream.list)
    assert nr_of_performed() == 11
    assert len(res3[0]) == 11
    with caplog.at_level(logging.WARNING):
        caplog.clear()
        res4 = await cli.execute_cli_command('search is("bla") limit 2 | tag delete foo', stream.list)
        assert nr_of_performed() == 2
        assert len(res4[0]) == 2
        # make sure that 2 warnings are emitted
        assert len(caplog.records) == 2
        for res in caplog.records:
            assert res.message.startswith("Update not reflected in db. Wait until next collector run.")
    # tag updates can be put into background
    res6 = await cli.execute_cli_command('json ["root", "collector"] | tag update --nowait foo bla', stream.list)
    assert cli.dependencies.forked_tasks.qsize() == 2
    for res in res6[0]:
        # in this case a message with the task id is emitted
        assert res.startswith("Spawned WorkerTask tag:")  # type:ignore
        # and the real result is found when the forked task is awaited, which happens by the CLI reaper
        awaitable, info = await cli.dependencies.forked_tasks.get()
        assert (await awaitable)["id"] in ["root", "collector"]  # type:ignore


@pytest.mark.asyncio
async def test_kinds_command(cli: CLI, foo_model: Model) -> None:
    result = await cli.execute_cli_command("kind", stream.list)
    for kind in ["account", "bla", "child", "cloud", "parent", "region", "some_complex"]:
        assert kind in result[0]
    result = await cli.execute_cli_command("kind foo", stream.list)
    assert result[0][0] == {
        "name": "foo",
        "bases": ["base"],
        "properties": {
            "age": "duration",
            "ctime": "datetime",
            "identifier": "string",
            "kind": "string",
            "name": "string",
            "now_is": "datetime",
            "some_int": "int32",
            "some_string": "string",
        },
        "successors": ["bla"],
    }
    result = await cli.execute_cli_command("kind string", stream.list)
    assert result[0][0] == {"name": "string", "runtime_kind": "string"}
    result = await cli.execute_cli_command("kind -p reported.ctime", stream.list)
    assert result[0][0] == {
        "name": "datetime",
        "runtime_kind": "datetime",
        "appears_in": [
            "base",
            "foo",
            "bla",
            "cloud",
            "account",
            "region",
            "parent",
            "child",
            "some_complex",
            "predefined_properties",
        ],
    }
    with pytest.raises(Exception):
        await cli.execute_cli_command("kind foo bla bar", stream.list)


@pytest.mark.asyncio
async def test_sort_command(cli: CLI) -> None:
    async def identifiers(query: str) -> List[str]:
        result = await cli.execute_cli_command(query + " | dump", stream.list)
        return [r["reported"]["identifier"] for r in result[0]]

    id_wo = await identifiers("search is(bla) | sort identifier")
    id_asc = await identifiers("search is(bla) | sort identifier asc")
    id_desc = await identifiers("search is(bla) | sort identifier desc")
    id_kind = await identifiers("search is(bla) | sort identifier | sort kind")
    assert id_wo == id_asc
    assert id_wo == id_kind
    assert id_asc == list(reversed(id_desc))


@pytest.mark.asyncio
async def test_limit_command(cli: CLI) -> None:
    async def identifiers(query: str) -> List[str]:
        result = await cli.execute_cli_command(query + " | dump", stream.list)
        return [r["reported"]["identifier"] for r in result[0]]

    assert await identifiers("search is(bla) sort identifier | limit 1") == ["0_0"]
    assert await identifiers("search is(bla) sort identifier | limit 2") == ["0_0", "0_1"]
    assert await identifiers("search is(bla) sort identifier | limit 2, 2") == ["0_2", "0_3"]
    assert await identifiers("search is(bla) sort identifier | limit 10, 2") == ["1_0", "1_1"]
    assert await identifiers("search is(bla) sort identifier | limit 100, 2") == []


@pytest.mark.asyncio
async def test_list_command(cli: CLI) -> None:
    result = await cli.execute_cli_command('search is (foo) and identifier=="4" sort some_int | list', stream.list)
    assert len(result[0]) == 1
    assert result[0][0].startswith("kind=foo, identifier=4, some_int=0, age=")
    list_cmd = "list some_int as si, some_string"
    result = await cli.execute_cli_command(f'search is (foo) and identifier=="4" | {list_cmd}', stream.list)
    assert result[0] == ["si=0, some_string=hello"]

    # list is added automatically when no output renderer is defined and has the same behaviour as if it was given
    result = await cli.execute_cli_command('search is (foo) and identifier=="4" sort some_int', stream.list)
    assert result[0][0].startswith("kind=foo, identifier=4, some_int=0, age=")

    # List is using the correct type
    props = dict(id="test", a="a", b=True, c=False, d=None, e=12, f=1.234, reported={})
    result = await cli.execute_cli_command(f"json {json.dumps(props)}" " | list a,b,c,d,e,f", stream.list)
    assert result[0] == ["a=a, b=true, c=false, e=12, f=1.234"]

    # Queries that use the reported section, also interpret the list format in the reported section
    result = await cli.execute_cli_command(
        "search id(sub_root) limit 1 | list some_string, some_int, /metadata.node_id", stream.list
    )
    assert result[0] == ["some_string=hello, some_int=0, node_id=sub_root"]

    # List supports csv output
    props = dict(id="test", a="a", b=True, c=False, d=None, e=12, f=1.234, reported={})
    result = await cli.execute_cli_command(
        f"json {json.dumps(props)}" " | list --csv a,b,c,d,e,f,non_existent", stream.list
    )
    assert result[0] == ['"a","b","c","d","e","f","non_existent"', '"a",True,False,"",12,1.234,""']

    # List supports markdown output
    props = dict(id="test", a="a", b=True, c=False, d=None, e=12, f=1.234, reported={})
    result = await cli.execute_cli_command(
        f"json {json.dumps(props)}" " | list --markdown a,b,c,d,e,f,non_existent", stream.list
    )
    assert result[0] == [
        "|a|b   |c    |d   |e |f    |non_existent|",
        "|-|----|-----|----|--|-----|------------|",
        "|a|true|false|null|12|1.234|null        |",
    ]

    # List supports only markdown or csv, but not both at the same time
    props = dict(id="test", a="a", b=True, c=False, d=None, e=12, f=1.234, reported={})
    with pytest.raises(CLIParseError):
        await cli.execute_cli_command(f"json {json.dumps(props)}" " | list --csv --markdown", stream.list)


@pytest.mark.asyncio
async def test_jq_command(cli: CLI) -> None:
    ctx = CLIContext(env={"section": "reported"}, query=Query.by("test"))
    # .test -> .reported.test
    assert JqCommand.rewrite_props(".a,.b", ctx) == ".reported.a,.reported.b"
    # absolute paths are rewritten correctly
    assert JqCommand.rewrite_props("./reported", ctx) == ".reported"
    # object construction is supported
    assert JqCommand.rewrite_props("{a:.a, b:.b}", ctx) == "{a:.reported.a, b:.reported.b}"
    # no replacement after pipe
    assert JqCommand.rewrite_props("map(.color) | {a:.a, b:.b}", ctx) == "map(.reported.color) | {a:.a, b:.b}"

    assert (
        JqCommand.rewrite_props(".pod_status.container_statuses[].image_id", ctx)
        == ".reported.pod_status.container_statuses[].image_id"
    )

    result = await cli.execute_cli_command('json {"a":{"b":1}} | jq ".a.b"', stream.list)
    assert len(result[0]) == 1
    assert result[0][0] == 1

    # allow absolute paths as json path
    result = await cli.execute_cli_command('json {"id":"123", "reported":{"b":1}} | jq "./reported"', stream.list)
    assert result == [[{"b": 1}]]

    # jq .kind is rewritten as .reported.kind
    result = await cli.execute_cli_command("search is(foo) limit 2 | jq .kind", stream.list)
    assert result[0] == ["foo", "foo"]


@pytest.mark.asyncio
async def test_execute_search_command(cli: CLI) -> None:
    # regression test: this used to fail because the arg could not be parsed
    await cli.execute_cli_command('execute_search (b= "0")', stream.list)


@pytest.mark.asyncio
async def test_aggregation_to_count_command(cli: CLI) -> None:
    r = await cli.execute_cli_command("search all | count kind", stream.list)
    assert set(r[0]) == {"graph_root: 1", "cloud: 1", "foo: 11", "bla: 100", "total matched: 113", "total unmatched: 0"}
    # exactly the same command as above (above search would be rewritten as this)
    r = await cli.execute_cli_command(
        "execute_search aggregate(reported.kind as name: sum(1) as count):all sort count asc | aggregate_to_count",
        stream.list,
    )
    assert set(r[0]) == {"graph_root: 1", "cloud: 1", "foo: 11", "bla: 100", "total matched: 113", "total unmatched: 0"}


@pytest.mark.skipif(not_in_path("arangodump"), reason="requires arangodump to be in path")
@pytest.mark.asyncio
async def test_system_backup_command(cli: CLI) -> None:
    async def check_backup(res: Stream) -> None:
        async with res.stream() as streamer:
            only_one = True
            async for s in streamer:
                path = FilePath.from_path(s)
                assert path.local.exists()
                # backup should have size between 30k and 1500k (adjust size if necessary)
                assert 30000 < path.local.stat().st_size < 1500000
                assert only_one
                only_one = False

    await cli.execute_cli_command("system backup create", check_backup)


@pytest.mark.asyncio
async def test_system_info_command(cli: CLI) -> None:
    info = AccessJson.wrap_object((await cli.execute_cli_command("system info", stream.list))[0][0])
    assert info.version == version()
    assert info.name == "resotocore"
    assert info.cpus > 0


@pytest.mark.skipif(not_in_path("arangodump", "arangorestore"), reason="requires arangodump and arangorestore")
@pytest.mark.asyncio
async def test_system_restore_command(cli: CLI, tmp_directory: str) -> None:
    backup = os.path.join(tmp_directory, "backup")

    async def move_backup(res: Stream) -> None:
        async with res.stream() as streamer:
            async for s in streamer:
                path = FilePath.from_path(s)
                os.rename(path.local, backup)

    await cli.execute_cli_command("system backup create", move_backup)
    ctx = CLIContext(uploaded_files={"backup": backup})
    restore = await cli.execute_cli_command(f"BACKUP_NO_SYS_EXIT=true system backup restore {backup}", stream.list, ctx)
    assert restore == [
        [
            "Database has been restored successfully!",
            "Since all data has changed in the database eventually, this service needs to be restarted!",
        ]
    ]


@pytest.mark.asyncio
async def test_configs_command(cli: CLI, tmp_directory: str) -> None:
    config_file = os.path.join(tmp_directory, "config.yml")

    async def check_file_is_yaml(res: Stream) -> None:
        async with res.stream() as streamer:
            async for s in streamer:
                with open(s, "r") as file:
                    yaml.safe_load(file.read())

    # create a new config entry
    create_result = await cli.execute_cli_command("configs set test_config t1=1, t2=2, t3=3 ", stream.list)
    assert create_result[0][0] == "t1: 1\nt2: 2\nt3: 3\n"
    # show the entry - should be the same as the created one
    show_result = await cli.execute_cli_command("configs show test_config", stream.list)
    assert show_result[0][0] == "t1: 1\nt2: 2\nt3: 3\n"
    # list all configs: only one is defined
    list_result = await cli.execute_cli_command("configs list", stream.list)
    assert list_result[0] == ["test_config"]

    # copy the config
    await cli.execute_cli_command("configs copy test_config test_config_copy", stream.list)
    list_result = await cli.execute_cli_command("configs list", stream.list)
    assert list_result[0] == ["test_config", "test_config_copy"]

    # edit the config: will make the config available as file
    await cli.execute_cli_command("configs edit test_config", check_file_is_yaml)
    # update the config
    update_doc = "a: '1'\nb: 2\nc: true\nd: null\n"
    with open(config_file, "w") as file:
        file.write(update_doc)
    ctx = CLIContext(uploaded_files={"config.yaml": config_file})
    update_result = await cli.execute_cli_command(f"configs update test_config {config_file}", stream.list, ctx)
    assert update_result == [[]]
    # show the entry - should be the same as the created one
    show_updated_result = await cli.execute_cli_command("configs show test_config", stream.list)
    assert show_updated_result[0][0] == update_doc
    # write a env var substitution to the config
    env_var_update = "foo: $(FOO)\n"
    with open(config_file, "w") as file:
        file.write(env_var_update)
    ctx = CLIContext(uploaded_files={"config.yaml": config_file})
    update_result = await cli.execute_cli_command(f"configs update test_config {config_file}", stream.list, ctx)
    # provide the env var
    os.environ["FOO"] = "bar"
    # check the configs: the env var should stay here and not be resolved when the user views the config
    show_updated_result = await cli.execute_cli_command("configs show test_config", stream.list)
    assert show_updated_result[0][0] == env_var_update


@pytest.mark.asyncio
async def test_templates_command(cli: CLI) -> None:
    result = await cli.execute_cli_command("templates test kind=volume is({{kind}})", stream.list)
    assert result == [["is(volume)"]]
    result = await cli.execute_cli_command("templates add filter_kind is({{kind}})", stream.list)
    assert result == [["Template filter_kind added to the search library.\nis({{kind}})"]]
    result = await cli.execute_cli_command("templates", stream.list)
    assert result == [["filter_kind: is({{kind}})"]]
    result = await cli.execute_cli_command("templates filter_kind", stream.list)
    assert result == [["is({{kind}})"]]
    result = await cli.execute_cli_command("templates delete filter_kind", stream.list)
    assert result == [["Template filter_kind deleted from the search library."]]


@pytest.mark.asyncio
async def test_write_command(cli: CLI) -> None:
    async def check_file(res: Stream, check_content: Optional[str] = None) -> None:
        async with res.stream() as streamer:
            only_one = True
            async for s in streamer:
                fp = FilePath.from_path(s)
                assert fp.local.exists() and fp.local.is_file()
                assert 1 < fp.local.stat().st_size < 100000
                assert fp.user.name.startswith("write_test")
                assert only_one
                only_one = False
                if check_content:
                    with open(fp.local, "r") as file:
                        data = file.read()
                        assert data == check_content

    # result can be read as json
    await cli.execute_cli_command("search all limit 3 | format --json | write write_test.json ", check_file)
    # result can be read as yaml
    await cli.execute_cli_command("search all limit 3 | format --yaml | write write_test.yaml ", check_file)
    # throw an exception
    with pytest.raises(Exception):
        await cli.execute_cli_command("echo hello | write", stream.list)  # missing filename
    # write enforces unescaped output.
    env = {"now": utc_str()}  # fix the time, so that replacements will stay equal
    truecolor = CLIContext(console_renderer=ConsoleRenderer(80, 25, ConsoleColorSystem.truecolor, True), env=env)
    monochrome = CLIContext(console_renderer=ConsoleRenderer.default_renderer(), env=env)
    # Make sure, that the truecolor output is different from monochrome output
    mono_out = await cli.execute_cli_command("help", stream.list, monochrome)
    assert await cli.execute_cli_command("help", stream.list, truecolor) != mono_out
    # We expect the content of the written file to contain monochrome output.
    assert await cli.execute_cli_command(
        "help | write write_test.txt", partial(check_file, check_content="".join(mono_out[0]) + "\n"), truecolor
    )


@pytest.mark.asyncio
async def test_http_command(cli: CLI, echo_http_server: Tuple[int, List[Tuple[Request, Json]]]) -> None:
    port, requests = echo_http_server

    def test_arg(
        arg_str: str,
        method: Optional[str] = None,
        url: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        timeout: Optional[ClientTimeout] = None,
        compress: Optional[bool] = None,
    ) -> None:
        def test_if_set(prop: Any, value: Any) -> None:
            if prop is not None:
                assert prop == value, f"{prop} is not {value}"

        arg = HttpCommand.parse_args("https", arg_str)
        test_if_set(method, arg.method)
        test_if_set(url, arg.url)
        test_if_set(headers, arg.headers)
        test_if_set(params, arg.params)
        test_if_set(compress, arg.compress)
        test_if_set(timeout, arg.timeout)

    test_arg(":123", "POST", "https://localhost:123", {}, {}, ClientTimeout(30), False)
    test_arg("GET :123", "GET", "https://localhost:123")
    test_arg("://foo:123", "POST", "https://foo:123")
    test_arg("foo:123/bla", "POST", "https://foo:123/bla")
    test_arg("foo:123/bla", "POST", "https://foo:123/bla")
    test_arg("foo/bla", "POST", "https://foo/bla")
    test_arg(
        '--compress --timeout 24 POST :123 "hdr1: test" qp==123  hdr2:fest "qp2 == 321"',
        headers={"hdr1": "test", "hdr2": "fest"},
        params={"qp": "123", "qp2": "321"},
        compress=True,
        timeout=ClientTimeout(24),
    )

    # take 3 instance of type bla and send it to the echo server
    result = await cli.execute_cli_command(f"search is(bla) limit 3 | http :{port}/test", stream.list)
    # one line is returned to the user with a summary of the response types.
    assert result == [["3 requests with status 200 sent."]]
    # make sure all 3 requests have been received - the body is the complete json node
    assert len(requests) == 3
    for ar in (AccessJson(content) for _, content in requests):
        assert is_node(ar)
        assert ar.reported.kind == "bla"

    # failing requests are retried
    requests.clear()
    await cli.execute_cli_command(f"search is(bla) limit 1 | http --backoff-base 0.001 :{port}/fail", stream.list)
    # 1 request + 3 retries => 4 requests
    assert len(requests) == 4


@pytest.mark.asyncio
async def test_jira_alias(cli: CLI, echo_http_server: Tuple[int, List[Tuple[Request, Json]]]) -> None:
    port, requests = echo_http_server
    result = await cli.execute_cli_command(
        f'search is(bla) | jira --url "http://localhost:{port}/success" --title test --message "test message" --username test --token test --project_id 10000 --reporter_id test',
        stream.list,
    )
    assert result == [["1 requests with status 200 sent."]]
    assert len(requests) == 1
    print(requests[0][1])
    assert requests[0][1] == {
        "fields": {
            "summary": "test",
            "issuetype": {"id": "10001"},
            "project": {"id": "10000"},
            "description": "test message\n\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\nbla: yes or no\n... (results truncated)\n\nIssue created by Resoto",
            "reporter": {"id": "test"},
            "labels": ["created-by-resoto"],
        }
    }


@pytest.mark.asyncio
async def test_pagerduty_alias(cli: CLI, echo_http_server: Tuple[int, List[Tuple[Request, Json]]]) -> None:
    port, requests = echo_http_server
    result = await cli.execute_cli_command(
        f'search id(0_0) | pagerduty --webhook-url "http://localhost:{port}/success" --summary test --routing-key 123 --dedup-key 234',
        stream.list,
    )
    assert result == [["1 requests with status 200 sent."]]
    assert len(requests) == 1
    response = requests[0][1]
    # override timestamp
    assert response["payload"]["timestamp"] is not None
    response["payload"]["timestamp"] = "2023-02-10T15:03:33Z"
    assert requests[0][1] == {
        "payload": {
            "summary": "test",
            "timestamp": "2023-02-10T15:03:33Z",
            "source": "Resoto",
            "severity": "warning",
            "component": "Resoto",
            "custom_details": {
                "collector": {"sub_root": {"no-region": {"0_0": {"id": None, "name": "yes or no", "kind": "bla"}}}}
            },
        },
        "routing_key": "123",
        "dedup_key": "234",
        "images": [
            {
                "src": "https://cdn.some.engineering/assets/resoto-illustrations/small/resoto-alert.png",
                "href": "https://resoto.com/",
                "alt": "Resoto Home Page",
            }
        ],
        "links": [],
        "event_action": "trigger",
        "client": "Resoto Service",
        "client_url": "https://resoto.com",
    }


@pytest.mark.asyncio
async def test_welcome(cli: CLI) -> None:
    ctx = CLIContext(console_renderer=ConsoleRenderer.default_renderer())
    result = await cli.execute_cli_command(f"welcome", stream.list, ctx)
    assert "Resoto" in result[0][0]


@pytest.mark.asyncio
async def test_tip_of_the_day(cli: CLI) -> None:
    ctx = CLIContext(console_renderer=ConsoleRenderer.default_renderer())
    result = await cli.execute_cli_command(f"totd", stream.list, ctx)
    assert generic_tips[0].command_line in result[0][0]


@pytest.mark.asyncio
async def test_certificate(cli: CLI) -> None:
    result = await cli.execute_cli_command(
        f"certificate create --common-name foo.resoto.com --dns-names bla --ip-addresses 1.2.3.4 --days-valid 1",
        stream.list,
    )
    # will create 2 files
    assert len(result[0]) == 2
    assert [a.rsplit("/")[-1] for a in result[0]] == ["foo.resoto.com.key", "foo.resoto.com.crt"]


@pytest.mark.asyncio
async def test_execute_task(cli: CLI) -> None:
    # translate a custom command to an alias template
    command = WorkerCustomCommand("name", "info", {"a": "b"}, "description").to_template()
    assert command.name == "name"
    assert command.info == "info"
    assert command.description == "description"
    assert command.args_description == {"a": "b"}
    assert command.template == "execute-task --no-node-result --command 'name' --arg '{{args}}'"

    # execute-task in source position
    source_result = await cli.execute_cli_command(
        f'execute-task --command success_task --arg "--foo bla test"', stream.list
    )
    assert len(source_result[0]) == 1
    assert source_result[0] == [{"result": "done!"}]

    # execute task in flow position: every incoming node creates a new task
    flow_result = await cli.execute_cli_command(
        f'search all limit 3 | execute-task --command success_task --arg "--t {{id}}"', stream.list
    )
    assert len(flow_result[0]) == 3


@pytest.mark.asyncio
async def test_history(cli: CLI, filled_graph_db: ArangoGraphDB) -> None:
    async def history_count(cmd: str) -> int:
        result = await cli.execute_cli_command(cmd, stream.list)
        return len(result[0])

    now = utc()
    five_min_ago = utc_str(now - timedelta(minutes=5))
    five_min_later = utc_str(now + timedelta(minutes=5))
    assert await history_count("history") == 113  # 112 inserts and 1 update for the filled graph db
    assert await history_count(f"history --after {five_min_ago}") == 113
    assert await history_count(f"history --after 5m") == 113
    assert await history_count(f"history --after {five_min_later}") == 0
    assert await history_count(f"history --before {five_min_ago}") == 0
    assert await history_count(f"history --before 5m") == 0
    assert await history_count(f"history --change node_created") == 112
    assert await history_count(f"history --change node_updated") == 1
    assert await history_count(f"history --change node_deleted") == 0
    assert await history_count(f"history --change node_deleted") == 0
    assert await history_count(f"history is(foo)") == 11
    # combine all selectors
    assert await history_count(f"history --after 5m --before {five_min_later} --change node_created is(foo)") == 11


@pytest.mark.asyncio
async def test_aggregate(dependencies: Dependencies) -> None:
    in_stream = stream.iterate(
        [{"a": 1, "b": 1, "c": 1}, {"a": 2, "b": 1, "c": 1}, {"a": 3, "b": 2, "c": 1}, {"a": 4, "b": 2, "c": 1}]
    )

    async def aggregate(agg_str: str) -> List[Json]:
        res = AggregateCommand(dependencies).parse(agg_str)
        async with (await res.flow(in_stream)).stream() as flow:
            return [s async for s in flow]

    assert await aggregate("b as bla, c, r.d.f.name: sum(1) as count, min(a) as min, max(a) as max") == [
        {"group": {"bla": 1, "c": 1, "r.d.f.name": None}, "count": 2, "min": 1, "max": 2},
        {"group": {"bla": 2, "c": 1, "r.d.f.name": None}, "count": 2, "min": 3, "max": 4},
    ]
    assert await aggregate("b as nb, c as nc: avg(a) as a, avg(b) as b, avg(c) as c") == [
        {"group": {"nb": 1, "nc": 1}, "a": 1.5, "b": 1, "c": 1},
        {"group": {"nb": 2, "nc": 1}, "a": 3.5, "b": 2, "c": 1},
    ]
    assert await aggregate("b: sum(1) as count") == [
        {"group": {"b": 1}, "count": 2},
        {"group": {"b": 2}, "count": 2},
    ]
    assert await aggregate('"{b}_{c}_{does_not_exist}" as name: sum(1) as count') == [
        {"group": {"name": "1_1_null"}, "count": 2},
        {"group": {"name": "2_1_null"}, "count": 2},
    ]


@pytest.mark.asyncio
async def test_report(cli: CLI, inspector_service: Inspector, test_benchmark: BenchmarkConfig) -> None:
    T = TypeVar("T")

    async def execute(cmd: str, _: Type[T]) -> List[T]:
        result = await cli.execute_cli_command(cmd, stream.list)
        return cast(List[T], result[0])

    # all benchmarks are listed
    assert "test" in await execute("report benchmark list", str)
    assert "test" in await execute("report benchmarks list", str)
    # all checks are listed
    assert "test_test_some_check" in await execute("report check list", str)
    assert "test_test_some_check" in await execute("report checks list", str)
    # the whole benchmark is printed to the user
    assert "Test section" in (await execute("report benchmark show test", Json))[0]
    # a single check is executed and produces a benchmark result: benchmark_node, check_node, edge == 3
    assert len((await execute("report check run test_test_some_check | dump", Json))) == 3
    # without output transformer, a markdown report is generated
    assert len((await execute("report check run test_test_some_check", str))) == 1
    # execute the test benchmark
    assert len((await execute("report benchmark run test | dump", Json))) == 9
    assert len((await execute("report benchmark run test --only-failing | dump", Json))) == 9
    assert len((await execute("report benchmark run test --severity critical | dump", Json))) == 0


@pytest.mark.asyncio
async def test_apps(cli: CLI, package_manager: PackageManager, infra_apps_runtime: Runtime, tmp_directory: str) -> None:
    T = TypeVar("T")

    async def execute(cmd: str, _: Type[T]) -> List[T]:
        result = await cli.execute_cli_command(cmd, stream.list)
        return cast(List[T], result[0])

    async def check_file_is_yaml(res: Stream) -> None:
        async with res.stream() as streamer:
            async for s in streamer:
                with open(s, "r") as file:
                    yaml.safe_load(file.read())

    # install a package
    assert "installed successfully" in (await execute("apps install cleanup-untagged", str))[0]
    manifest = await package_manager.get_manifest(InfraAppName("cleanup-untagged"))
    assert manifest is not None
    assert manifest.name == "cleanup-untagged"
    # install discord app
    assert "installed successfully" in (await execute("apps install discord", str))[0]

    # info about the app
    info_json = (await execute("apps info cleanup-untagged", Json))[0]
    assert info_json["name"] == "cleanup-untagged"

    # run the app
    result = await execute("apps run cleanup-untagged --dry-run", str)
    assert result[0].startswith("search /metadata.protected == false and /metadata.phantom")

    # run the app with stdin
    result = await execute("echo foo | apps run cleanup-untagged --dry-run", str)
    assert result[0].startswith("search /metadata.protected == false and /metadata.phantom")

    # run discord app with stdin
    result = await execute("search is(graph_root) | apps run discord --title foo --dry-run", str)
    assert "http POST https://discordapp.com" in result[0]
    await execute("apps uninstall discord", str)

    # update the app
    assert (
        "App cleanup-untagged updated sucessfully to the latest version"
        in (await execute("apps update cleanup-untagged", str))[0]
    )

    # update all apps
    assert (
        "App cleanup-untagged updated sucessfully to the latest version"
        in (await execute("apps update cleanup-untagged", str))[0]
    )

    # edit the manifest: will make the manifest available as file
    manifest_file = os.path.join(tmp_directory, "manifest.yml")
    old_manifest = await cli.dependencies.infra_apps_package_manager.get_manifest(InfraAppName("cleanup-untagged"))
    assert old_manifest is not None
    await cli.execute_cli_command("apps edit cleanup-untagged", check_file_is_yaml)
    # update the manifest
    updated_manifest = evolve(old_manifest, version="42")
    updated_manifest_str = yaml.dump(to_js(updated_manifest))
    with open(manifest_file, "w", encoding="utf-8") as file:
        file.write(updated_manifest_str)
    ctx = CLIContext(uploaded_files={"manifest.yaml": manifest_file})
    update_result = await cli.execute_cli_command(f"apps update cleanup-untagged {manifest_file}", stream.list, ctx)
    assert update_result == [[]]
    # show the manifest - should be the same as the created one
    updated_result = await cli.dependencies.infra_apps_package_manager.get_manifest(InfraAppName("cleanup-untagged"))
    assert updated_result == updated_manifest

    # list all apps
    result = await execute("apps list", str)
    assert result == ["cleanup-untagged"]

    # uninstall the app
    await execute("apps uninstall cleanup-untagged", str)
    result = await execute("apps list", str)
    assert result == []


@pytest.mark.asyncio
async def test_user(cli: CLI) -> None:
    async def execute(cmd: str) -> List[JsonElement]:
        all_results = await cli.execute_cli_command(cmd, stream.list)
        return all_results[0]  # type: ignore

    # remove all existing users
    await cli.dependencies.config_handler.delete_config(UsersConfigId)

    # create new user
    result = await execute('user add john@test.de --fullname "John Doe" --password test --role test')
    assert result == [{"email": "john@test.de", "fullname": "John Doe", "roles": ["test"]}]

    # get user
    result = await execute("user show john@test.de")
    assert result == [{"email": "john@test.de", "fullname": "John Doe", "roles": ["test"]}]

    # add role to user
    result = await execute("user role add john@test.de test2")
    roles = set(result[0]["roles"])  # type: ignore
    assert roles == {"test", "test2"}

    # remove role from user
    result = await execute("user role delete john@test.de test2")
    assert result == [{"email": "john@test.de", "fullname": "John Doe", "roles": ["test"]}]

    # Change password
    result = await execute("user password john@test.de bombproof")
    assert result == ["Password for john@test.de updated"]

    # create another user
    result = await execute('user add jane@test.de --fullname "Jane Doe" --password test --role admin')
    assert result == [{"email": "jane@test.de", "fullname": "Jane Doe", "roles": ["admin"]}]

    # list users
    result = await execute("user list")
    assert result == ["john@test.de", "jane@test.de"]

    # delete user
    result = await execute("user delete john@test.de")
    assert result == ["User john@test.de deleted"]

    # list users
    result = await execute("user list")
    assert result == ["jane@test.de"]


@pytest.mark.asyncio
async def test_graph(cli: CLI, graph_manager: GraphManager, tmp_directory: str) -> None:
    T = TypeVar("T")

    await graph_manager.delete(GraphName("graphtest2"))
    await graph_manager.delete(GraphName("graphtest_import"))

    async def execute(cmd: str, _: Type[T]) -> List[T]:
        result = await cli.execute_cli_command(cmd, stream.list)
        return cast(List[T], result[0])

    # cleanup everything
    for graph in await graph_manager.list(None):
        await graph_manager.delete(graph)

    # create a graph
    await graph_manager.db_access.create_graph(GraphName("graphtest"))
    await graph_manager.db_access.create_graph(GraphName("ns"))

    # list all graphs
    graphs = await execute("graph list", str)
    assert set(graphs) == {"graphtest", "ns"}

    # list via regex
    graphs = await execute("graph list .*apht.*", str)
    assert set(graphs) == {"graphtest"}

    # copy a graph
    await execute("graph copy graphtest graphtest2", str)
    graph_names = await graph_manager.list(None)
    assert set(graph_names) == {"ns", "graphtest", "graphtest2"}

    # copy to the existing graph without --force
    with pytest.raises(Exception):
        await execute("graph copy graphtest graphtest2", str)

    # copy to the existing graph with --force
    await execute("graph copy graphtest graphtest2 --force", str)
    graph_names = await graph_manager.list(None)
    assert set(graph_names) == {"ns", "graphtest", "graphtest2"}

    # implicitly copy the current graph to the new one
    await execute("graph copy graphtest3", str)
    graph_names = await graph_manager.list(None)
    assert set(graph_names) == {"ns", "graphtest", "graphtest2", "graphtest3"}

    # make a snapshot
    await execute("graph snapshot graphtest foo", str)
    snapshots = await graph_manager.list("snapshot.*")
    assert len(snapshots) == 1
    # implicitly use the current graph as a source
    await execute("graph snapshot foobar", str)
    snapshots = await graph_manager.list("snapshot.*")
    assert len(snapshots) == 2
    assert snapshots[0].startswith("snapshot")
    await asyncio.sleep(1.1)

    # search using a timestamp
    at_timestamp = await execute(f"search is(graph_root) --at {utc_str()}", str)
    assert at_timestamp == ["kind=graph_root, name=root"]
    # search using timedelta
    assert await execute("search is(graph_root) --at 0s", str) == ["kind=graph_root, name=root"]

    for snapshot in snapshots:
        await graph_manager.delete(GraphName(snapshot))

    with pytest.raises(Exception):
        await execute(f"search is(graph_root) --at {utc_str()}", str)

    # delete a graph
    await execute("graph delete graphtest2", str)
    graph_names = await graph_manager.list(None)
    assert set(graph_names) == {"ns", "graphtest", "graphtest3"}

    dump = os.path.join(tmp_directory, "dump")

    async def move_dump(res: Stream) -> None:
        async with res.stream() as streamer:
            async for s in streamer:
                fp = FilePath.from_path(s)
                os.rename(fp.local, dump)

    # graph export works
    await cli.execute_cli_command("graph export graphtest dump", move_dump)

    ctx = CLIContext(uploaded_files={"dump": dump})

    # graph import works too
    await cli.execute_cli_command("graph import graphtest_import graphtest.backup", stream.list, ctx)
    assert await graph_manager.list(GraphName("graphtest_import")) == [GraphName("graphtest_import")]

    # clean up
    await graph_manager.delete(GraphName("graphtest3"))
    await graph_manager.delete(GraphName("graphtest_import"))


@pytest.mark.asyncio
async def test_db(cli: CLI) -> None:
    db_file = "test_db"

    async def sync_and_check(
        cmd: str,
        *,
        expected_table: Optional[Callable[[str, int], bool]] = None,
        expected_tables: Optional[Set[str]] = None,
        expected_table_count: Optional[int] = None,
    ) -> str:
        result: List[str] = []

        async def check(in_: Stream) -> None:
            async with in_.stream() as streamer:
                async for s in streamer:
                    path = FilePath.from_path(s)
                    # open sqlite database
                    conn = sqlite3.connect(path.local)
                    c = conn.cursor()
                    tables = {
                        row[0] for row in c.execute("SELECT tbl_name FROM sqlite_master WHERE type='table'").fetchall()
                    }
                    if expected_tables is not None:
                        assert tables == expected_tables
                    if expected_table_count is not None:
                        assert len(tables) == expected_table_count
                    if expected_table is not None:
                        for table in tables:
                            count = c.execute(f"SELECT COUNT(*) FROM {table}").fetchone()[0]
                            assert expected_table(table, count), f"Table {table} has {count} rows"
                    c.close()
                    conn.close()
                    result.append(s)

        await cli.execute_cli_command(cmd, check)
        assert len(result) == 1
        return result[0]

    # search | db sync
    await sync_and_check(
        f"search --with-edges is(foo) -[0:1]-> | db sync sqlite --database {db_file}",
        expected_table=lambda table, count: count > 0 if not table.startswith("link_") else True,
        expected_tables={"foo", "bla", "link_bla_bla", "link_foo_bla"},
    )

    # db sync synchronizes the whole graph
    await sync_and_check(
        f"db sync sqlite --database {db_file}",
        expected_table=lambda table, count: count > 0 if not table.startswith("link_") else True,
        expected_tables={"foo", "bla", "link_bla_bla", "link_foo_bla"},
    )

    # db sync with complete schema synchronizes the whole graph and created tables for all kinds even if they are empty
    await sync_and_check(f"db sync sqlite --complete-schema --database {db_file}", expected_table_count=11)

    # support write after db sync (not required for simple files, but we want to support s3 etc. in the future)
    path_result = await sync_and_check(f"db sync sqlite --database {db_file} | write out.db")
    assert FilePath.from_path(path_result).user.name == "out.db"

    # search with aggregation does not export anything
    with pytest.raises(Exception):
        await sync_and_check(f"search all | aggregate kind:sum(1) | db sync sqlite --database foo")

    # define all parameters and check the connection string
    with pytest.raises(Exception) as ex:
        await sync_and_check(
            f"db sync sqlite --database db --host bla --port 1234 --user test --password check --arg foo=bla foo2=bla2",
            expected_table_count=11,
        )
    assert "sqlite://test:check@bla:1234" in str(ex.value)
    assert "?foo=bla&foo2=bla2" in str(ex.value)

    # calling db without command will print the help
    db_res = await cli.execute_cli_command("db", stream.list)
    assert "Synchronizes data to an SQL database." in db_res[0][0]

    # make sure argsinfo is available
    assert "sync" in cli.direct_commands["db"].args_info()



import pytest

from resotocore.analytics import InMemoryEventSender, CoreEvent
from resotocore.config import ConfigHandler
from resotocore.db.system_data_db import SystemDataDb
from resotocore.ids import Email, Password
from resotocore.user import UserManagement, UsersConfigId


@pytest.mark.asyncio
async def test_user_creation(
    config_handler: ConfigHandler,
    user_management: UserManagement,
    event_sender: InMemoryEventSender,
    system_data_db: SystemDataDb,
) -> None:
    await config_handler.delete_config(UsersConfigId)  # clean slate
    assert await user_management.has_users() is False
    email = Email("test@some.engineering")
    password = Password("bombproof")
    user = await user_management.create_first_user("some company", "some name", email, password)
    # make sure user is created
    assert user.fullname == "some name"
    assert user.roles == {"admin"}
    # an analytics event should have been sent
    assert [event for event in event_sender.events if event.kind == CoreEvent.FirstUserCreated]
    # the company is set in the system data db
    assert (await system_data_db.info())["company"] == "some company"
    # can not create a second user
    with pytest.raises(AssertionError):
        await user_management.create_first_user("some company", "some name", email, password)
    # user login is possible
    assert await user_management.login(email, password) == user
    # wrong password is not accepted
    assert await user_management.login(email, Password("wrong password")) is None



from typing import NewType

TaskDescriptorId = NewType("TaskDescriptorId", str)
SubscriberId = NewType("SubscriberId", str)
TaskId = NewType("TaskId", str)
WorkerId = NewType("WorkerId", str)
ConfigId = NewType("ConfigId", str)
NodeId = NewType("NodeId", str)
InfraAppName = NewType("InfraAppName", str)
GraphName = NewType("GraphName", str)
Email = NewType("Email", str)
Password = NewType("Password", str)



import logging
import re
from abc import ABC, abstractmethod
from functools import reduce
from typing import Optional, List, Set, Callable, Dict, Iterator

from plantuml import PlantUML

from resotocore.async_extensions import run_async
from resotocore.db.db_access import DbAccess
from resotocore.ids import GraphName
from resotocore.model.model import Model, Kind, ComplexKind, Property
from resotocore.service import Service
from resotocore.types import EdgeType
from resotocore.util import exist

log = logging.getLogger(__name__)


class ModelHandler(ABC):
    @abstractmethod
    async def load_model(self, graph_name: GraphName, *, force: bool = False) -> Model:
        pass

    @abstractmethod
    async def uml_image(
        self,
        graph_name: GraphName,
        output: str = "svg",
        *,
        show_packages: Optional[List[str]] = None,
        hide_packages: Optional[List[str]] = None,
        with_inheritance: bool = True,
        with_base_classes: bool = False,
        with_subclasses: bool = False,
        dependency_edges: Optional[Set[EdgeType]] = None,
        with_predecessors: bool = False,
        with_successors: bool = False,
        with_properties: bool = True,
        link_classes: bool = False,
        only_aggregate_roots: bool = True,
        sort_props: bool = True,
    ) -> bytes:
        """
        Generate a PlantUML image of the model.

        :param graph_name: the name of the graph
        :param output: "svg" or "png"
        :param show_packages: regexp that matches packages to show
        :param hide_packages: regexp that matches packages to hide
        :param with_inheritance: draw inheritance relationship between classes
        :param with_base_classes: include base classes for all matching classes to show in the diagram
        :param with_subclasses: include subclasses for all matching classes to show in the diagram
        :param dependency_edges: draw dependency edges of given edge type between classes
        :param with_predecessors: include predecessors for all matching classes to show in the diagram
        :param with_successors: include successors for all matching classes to show in the diagram
        :param with_properties: include properties for all matching classes to show in the diagram
        :param link_classes: add anchor links to all classes
        :param only_aggregate_roots: if the list of classes should be filtered for aggregate roots
        :param sort_props: if the properties should be sorted by name
        :return: the generated image
        """

    @abstractmethod
    async def update_model(self, graph_name: GraphName, kinds: List[Kind]) -> Model:
        pass


PlantUmlAttrs = (
    "hide empty members\n"
    "skinparam ArrowColor #ffaf37\n"
    "skinparam ArrowFontColor #ffaf37\n"
    "skinparam ArrowFontName Helvetica\n"
    "skinparam ArrowThickness 2\n"
    "skinparam BackgroundColor transparent\n"
    "skinparam ClassAttributeFontColor #d9b8ff\n"
    "skinparam ClassBackgroundColor #3d176e\n"
    "skinparam ClassBorderColor #000d19\n"
    "skinparam ClassFontColor #d9b8ff\n"
    "skinparam ClassFontName Helvetica\n"
    "skinparam ClassFontSize 17\n"
    "skinparam NoteBackgroundColor #d9b8ff\n"
    "skinparam NoteBorderColor #000d19\n"
    "skinparam NoteFontColor #3d176e\n"
    "skinparam NoteFontName Helvetica\n"
    "skinparam Padding 5\n"
    "skinparam RoundCorner 5\n"
    "skinparam Shadowing false\n"
    "skinparam stereotypeCBackgroundColor #e98df7\n"
    "skinparam stereotypeIBackgroundColor #e98df7\n"
)


class ModelHandlerDB(ModelHandler, Service):
    def __init__(self, db_access: DbAccess, plantuml_server: str = "https://plantuml.resoto.org"):
        self.db_access = db_access
        self.plantuml_server = plantuml_server
        self.__loaded_model: Dict[GraphName, Model] = {}
        self.default_legacy_graph_name = GraphName("resoto")

    async def load_model(self, graph_name: GraphName, *, force: bool = False) -> Model:
        if not force and (model := self.__loaded_model.get(graph_name)) is not None:
            return model
        else:
            graph_model_db = await self.db_access.get_graph_model_db(graph_name)
            model = Model.from_kinds([kind async for kind in graph_model_db.all()])
            self.__loaded_model[graph_name] = model
            return model

    async def uml_image(
        self,
        graph_name: GraphName,
        output: str = "svg",
        *,
        show_packages: Optional[List[str]] = None,
        hide_packages: Optional[List[str]] = None,
        with_inheritance: bool = True,
        with_base_classes: bool = False,
        with_subclasses: bool = False,
        dependency_edges: Optional[Set[EdgeType]] = None,
        with_predecessors: bool = False,
        with_successors: bool = False,
        with_properties: bool = True,
        link_classes: bool = False,
        only_aggregate_roots: bool = True,
        sort_props: bool = True,
    ) -> bytes:
        allowed_edge_types: Set[EdgeType] = dependency_edges or set()
        assert output in ("svg", "png", "puml"), "Only svg, png and puml is supported!"
        model = await self.load_model(graph_name)
        graph = model.graph()
        show = [re.compile(s) for s in show_packages] if show_packages else None
        hide = [re.compile(s) for s in hide_packages] if hide_packages else None
        visited_complex: Set[str] = set()

        def not_hidden(key: str) -> bool:
            k: Kind = graph.nodes[key]["data"]
            return not (hide and exist(lambda r: r.fullmatch(k.fqn), hide))

        def node_visible(key: str) -> bool:
            k: Kind = graph.nodes[key]["data"]
            ar_visible = not only_aggregate_roots or getattr(k, "aggregate_root", False)
            if hide and exist(lambda r: r.fullmatch(k.fqn), hide):
                return False
            if show is None:
                return ar_visible
            else:
                return ar_visible and exist(lambda r: r.fullmatch(k.fqn), show)

        def class_node(cpx: ComplexKind) -> str:
            sth = {k.prop.name: k for k in cpx.synthetic_props() if len(k.path.path) == 1}

            def kind_name(p: Property) -> str:
                return (sth[p.name].simple_kind.runtime_kind if p.name in sth else p.kind) if p.synthetic else p.kind

            cpx_props = sorted(cpx.properties, key=lambda p: p.name) if sort_props else cpx.properties
            props = "\n".join([f"**{p.name}**: {kind_name(p)}" for p in cpx_props]) if with_properties else ""
            link = f" [[#{cpx.fqn}]]" if link_classes else ""
            return f"class {cpx.fqn}{link} {{\n{props}\n}}"

        def descendants(cpx: ComplexKind) -> Set[str]:
            return {kind.fqn for kind in model.complex_kinds() if cpx.fqn in kind.kind_hierarchy()}

        def predecessors(cpx: ComplexKind) -> Set[str]:
            return {
                kind.fqn
                for kind in model.complex_kinds()
                for et in allowed_edge_types
                if cpx.fqn in kind.successor_kinds.get(et, [])
            }

        def successors(cpx: ComplexKind) -> Set[str]:
            return (
                {kind for et in allowed_edge_types for kind in cpx.successor_kinds.get(et, [])}
                if isinstance(cpx, ComplexKind)
                else set()
            )

        def complex_property_kinds(cpx: ComplexKind) -> Set[str]:
            if cpx.fqn in visited_complex:
                return set()
            visited_complex.add(cpx.fqn)
            result: Set[str] = set()
            for _, k in cpx.property_with_kinds():
                for cpl in k.nested_complex_kinds():
                    result.add(cpl.fqn)
                    result.update(complex_property_kinds(cpl))
            return result

        visible_kinds = [node["data"] for nid, node in graph.nodes(data=True) if node_visible(nid)]
        visible = {v.fqn for v in visible_kinds}

        def add_visible(fn: Callable[[ComplexKind], Set[str]]) -> None:
            selected: Set[str] = reduce(lambda res, cpl: res.union(fn(cpl)), visible_kinds, set())
            visible.update(n for n in selected if not_hidden(n))

        if with_base_classes:
            add_visible(lambda cpl: cpl.kind_hierarchy())
        if with_subclasses:
            add_visible(descendants)
        if with_predecessors:
            add_visible(predecessors)
        if with_successors:
            add_visible(successors)
        if with_properties:
            add_visible(complex_property_kinds)

        visible_nodes: Iterator[ComplexKind] = (node["data"] for nid, node in graph.nodes(data=True) if nid in visible)
        nodes = "\n".join(class_node(node) for node in sorted(visible_nodes, key=lambda n: n.fqn))
        edges = ""
        for fr, to, data in sorted(graph.edges(data=True), key=lambda e: (e[0], e[1])):
            if fr in visible and to in visible:
                if with_inheritance and data["type"] == "inheritance":
                    edges += f"{to} <|--- {fr}\n"
                elif data["type"] == "successor" and data["edge_type"] in allowed_edge_types:
                    edges += f"{fr} -[#1A83AF]-> {to}\n"
                elif data["type"] == "property" and with_properties:
                    edges += f"{fr} --> {to}\n"

        puml = f"@startuml\n{PlantUmlAttrs}\n{nodes}\n{edges}\n@enduml"
        if output == "puml":
            return puml.encode("utf-8")
        else:
            plant_uml = PlantUML(f"{self.plantuml_server}/{output}/")
            return await run_async(plant_uml.processes, puml)

    async def update_model(self, graph_name: GraphName, kinds: List[Kind]) -> Model:
        # load existing model
        model = await self.load_model(graph_name)
        # make sure the update is valid
        updated = model.update_kinds(kinds)
        # store all updated kinds
        db = await self.db_access.get_graph_model_db(graph_name)
        await db.update_many(kinds)
        # unset loaded model
        self.__loaded_model[graph_name] = updated
        return updated



import json
import logging
import re
from collections import defaultdict
from textwrap import dedent
from typing import Union, List, Tuple, Any, Optional, Dict, Set

from arango.typings import Json
from attrs import evolve

from resotocore.constants import less_greater_then_operations as lgt_ops, arangodb_matches_null_ops
from resotocore.db import EstimatedSearchCost, EstimatedQueryCostRating as Rating
from resotocore.db.arangodb_functions import as_arangodb_function
from resotocore.db.model import QueryModel
from resotocore.model.graph_access import Section, Direction
from resotocore.model.model import SyntheticProperty, ResolvedProperty
from resotocore.model.resolve_in_graph import GraphResolver
from resotocore.query.model import (
    Predicate,
    IsTerm,
    Part,
    Term,
    CombinedTerm,
    FunctionTerm,
    Navigation,
    IdTerm,
    Aggregate,
    AllTerm,
    AggregateFunction,
    Sort,
    WithClause,
    AggregateVariableName,
    AggregateVariableCombined,
    NotTerm,
    MergeTerm,
    MergeQuery,
    Query,
    SortOrder,
    FulltextTerm,
    Limit,
    ContextTerm,
    WithUsage,
)
from resotocore.util import first, set_value_in_path, exist, utc_str
from resotolib.durations import duration_str

log = logging.getLogger(__name__)

# Those words are keywords in aql and need to be "escaped"
escape_aql_parts = {
    "collect",
    "filter",
    "for",
    "insert",
    "let",
    "limit",
    "remove",
    "replace",
    "return",
    "search",
    "sort",
    "update",
    "upsert",
    "window",
    "with",
}

allowed_first_merge_part = Part(AllTerm())
unset_props = json.dumps(["flat"])
# This list of delimiter is also used in the arango delimiter index.
# In case the definition is changed, also the index needs to change!
fulltext_delimiter = [" ", "_", "-", "@", ":", "/", "."]
fulltext_delimiter_regexp = re.compile("[" + "".join(re.escape(a) for a in fulltext_delimiter) + "]+")

# All resolved ancestors attributes have to be treated explicitly.
# Queries with /ancestors.kind.xxx have to be treated as merge query parameters.
ancestor_merges = {
    f"ancestors.{p.to_path[1]}" for r in GraphResolver.to_resolve for p in r.resolve if p.to_path[0] == "ancestors"
}

array_marker = re.compile(r"\[]|\[[*]]")
array_marker_in_path_regexp = re.compile(r"\[]|\[[*]](?=[.])")


def to_query(
    db: Any, query_model: QueryModel, with_edges: bool = False, from_collection: Optional[str] = None
) -> Tuple[str, Json]:
    count: Dict[str, int] = defaultdict(lambda: 0)
    query = query_model.query
    bind_vars: Json = {}
    start = from_collection or f"`{db.vertex_name}`"
    cursor, query_str = query_string(db, query, query_model, start, with_edges, bind_vars, count)
    return f"""{query_str} FOR result in {cursor} RETURN UNSET(result, {unset_props})""", bind_vars


def query_string(
    db: Any,
    query: Query,
    query_model: QueryModel,
    start_cursor: str,
    with_edges: bool,
    bind_vars: Json,
    counters: Dict[str, int],
    outer_merge: Optional[str] = None,
) -> Tuple[str, str]:
    # Note: the parts are maintained in reverse order
    query_parts = query.parts[::-1]
    model = query_model.model
    # combine merge names from the query as well as the default ancestor merge names
    merge_names: Set[str] = query_model.query.merge_names | ancestor_merges

    def next_counter(name: str) -> int:
        count = counters[name]
        counters[name] = count + 1
        return count

    def next_crs(name: str = "m") -> str:
        return f"{name}{next_counter(name)}"

    def next_bind_var_name() -> str:
        return f'b{next_counter("bind_vars")}'

    def prop_name_kind(
        path: str, context_path: Optional[str] = None
    ) -> Tuple[str, ResolvedProperty, Optional[str]]:  # prop_name, prop, merge_name
        local_path = f"{context_path}.{path}" if context_path else path
        merge_name = first(lambda name: local_path.startswith(name + "."), merge_names)
        # remove merge_name and section part (if existent) from the local_path
        lookup = Section.without_section(local_path[len(merge_name) + 1 :] if merge_name else local_path)  # noqa: E203
        resolved = model.property_by_path(lookup)

        def synthetic_path(synth: SyntheticProperty) -> str:
            before, after = local_path.rsplit(resolved.prop.name, 1)
            return f'{before}{".".join(synth.path)}{after}'

        def escape_part(path_part: str) -> str:
            return f"`{path_part}`" if path_part.lower() in escape_aql_parts else path_part

        prop_name = synthetic_path(resolved.prop.synthetic) if resolved.prop.synthetic else local_path

        # remove the context from the path
        if context_path and prop_name.startswith(context_path):
            prop_name = prop_name[len(context_path) + 1 :]
        # make sure the path does not contain any aql keywords
        prop_name = ".".join(escape_part(pn) for pn in prop_name.split("."))

        return prop_name, resolved, merge_name

    def aggregate(in_cursor: str, a: Aggregate) -> Tuple[str, str]:
        cursor = next_crs("agg")

        def var_name(n: Union[AggregateVariableName, AggregateVariableCombined]) -> str:
            def comb_name(cb: Union[str, AggregateVariableName]) -> str:
                return f'"{cb}"' if isinstance(cb, str) else f"{cursor}.{cb.name}"

            return (
                f"{cursor}.{n.name}"
                if isinstance(n, AggregateVariableName)
                else f'CONCAT({",".join(comb_name(cp) for cp in n.parts)})'
            )

        def func_term(fn: AggregateFunction) -> str:
            name = f"{cursor}.{fn.name}" if isinstance(fn.name, str) else str(fn.name)
            return f"{name} {fn.combined_ops()}" if fn.ops else name

        variables = ", ".join(f"var_{num}={var_name(v.name)}" for num, v in enumerate(a.group_by))
        funcs = ", ".join(f"fn_{num}={f.function}({func_term(f)})" for num, f in enumerate(a.group_func))
        agg_vars = ", ".join(f'"{v.get_as_name()}": var_{num}' for num, v in enumerate(a.group_by))
        agg_funcs = ", ".join(f'"{f.get_as_name()}": fn_{num}' for num, f in enumerate(a.group_func))
        group_result = f'"group":{{{agg_vars}}},' if a.group_by else ""
        aggregate_term = f"collect {variables} aggregate {funcs}"
        return_result = f"{{{group_result} {agg_funcs}}}"
        return "aggregated", f"LET aggregated = (for {cursor} in {in_cursor} {aggregate_term} RETURN {return_result})"

    def predicate(cursor: str, p: Predicate, context_path: Optional[str] = None) -> Tuple[Optional[str], str]:
        pre = ""
        extra = ""
        path = p.name

        # handle that the final property is an array: a.b.c[*]
        if "filter" in p.args:
            arr_filter = p.args["filter"]
            extra = f" {arr_filter} "
            path = p.name if p.name.endswith("[*]") or p.name.endswith("[]") else f"{p.name}[*]"
        elif p.name.endswith("[*]"):
            extra = " any "
            path = p.name
        elif p.name.endswith("[]"):
            extra = " any "
            path = p.name.replace("[]", "[*]")

        prop_name, prop, _ = prop_name_kind(path, context_path)

        # nested prop access needs to be unfolded via separate for loops: a.b[*].c[*].d
        ars = [a.lstrip(".") for a in array_marker_in_path_regexp.split(prop_name)]
        ars_stmts = []
        prop_name = ars.pop()
        for ar in ars:
            nxt_crs = next_crs("pre")
            ars_stmts.append(f"{nxt_crs}._internal!=true")
            # append an internal element to make sure this for loop always yields at least one result
            # this value will be filtered out explicitly later on.
            # example: group_ip_permissions[*].{ip_ranges[*].cidr_ip="0.0.0.0/0" or ipv6_ranges[*].cidr_ipv6="::/0"}
            # if there are no cidr_ip or cidr_ipv6 ranges, the nested for loops will not yield any results
            # the or condition will not be evaluated and the result will be empty
            # to avoid this, we add an internal element to the result set and filter it out later on
            pre += f" FOR {nxt_crs} IN APPEND(TO_ARRAY({cursor}.{ar}), {{_internal: true}})"
            cursor = nxt_crs

        bvn = next_bind_var_name()
        op = lgt_ops[p.op] if prop.simple_kind.reverse_order and p.op in lgt_ops else p.op
        if op in ["in", "not in"] and isinstance(p.value, list):
            bind_vars[bvn] = [prop.kind.coerce(a) for a in p.value]
        else:
            bind_vars[bvn] = prop.kind.coerce(p.value)
        var_name = f"{cursor}.{prop_name}"
        if op == "=~":  # use regex_test to do case-insensitive matching
            p_term = f"REGEX_TEST({var_name}, @{bvn}, true)"
        else:
            p_term = f"{var_name}{extra} {op} @{bvn}"
        if ars_stmts:
            p_term = f"({p_term} AND {' AND '.join(ars_stmts)})"
        # null check is required, since x<anything evaluates to true if x is null!
        return pre, f"({var_name}!=null and {p_term})" if op in arangodb_matches_null_ops else p_term

    def context_term(cursor: str, aep: ContextTerm, context_path: Optional[str] = None) -> Tuple[Optional[str], str]:
        predicate_statement = ""
        filter_statement = ""
        path_cursor = cursor
        context_path = f"{context_path}.{aep.name}" if context_path else aep.name
        # unfold only, if random access is required
        if "[]" in aep.name or "[*]" in aep.name:
            spath = array_marker.split(aep.name)
            # in case the array is defined on the last property, it can be ignored
            if spath[-1] == "":
                spath = spath[:-1]
            for ar in [a.lstrip(".") for a in spath]:
                nxt_crs = next_crs("pre")
                predicate_statement += f" FOR {nxt_crs} IN TO_ARRAY({path_cursor}.{ar})"
                path_cursor = nxt_crs
            ps, fs = term(path_cursor, aep.term, context_path)
        else:
            # no unfolding required, just use the current cursor
            # move the context path into the variable name, do not use any local path for rendering
            # (a.b.{c=1 and d=2}) ==> (a.b.c=1 and a.b.d=2)
            ps, fs = term(path_cursor, aep.term.change_variable(lambda x: f"{context_path}.{x}"))
        if ps:
            predicate_statement += ps
        if fs:
            filter_statement += fs
        return predicate_statement, filter_statement

    def with_id(cursor: str, t: IdTerm) -> str:
        bvn = next_bind_var_name()
        bind_vars[bvn] = t.id
        return f"{cursor}._key == @{bvn}"

    def is_term(cursor: str, t: IsTerm) -> str:
        is_results = []
        for kind in t.kinds:
            if kind not in model:
                raise AttributeError(f"Given kind does not exist: {kind}")
            bvn = next_bind_var_name()
            bind_vars[bvn] = kind
            is_results.append(f"@{bvn} IN {cursor}.kinds")
        is_result = " or ".join(is_results)
        return is_result if len(is_results) == 1 else f"({is_result})"

    def fulltext_term(cursor: str, t: FulltextTerm) -> str:
        # This fulltext filter can not take advantage of the fulltext search index.
        # Instead, we filter the resulting entry to match a regular expression derived from the term.
        # The flat property is used via a regexp search.
        bvn = next_bind_var_name()
        dl = fulltext_delimiter_regexp
        bind_vars[bvn] = dl.pattern.join(f"{re.escape(w)}" for w in dl.split(t.text))
        return f"REGEX_TEST({cursor}.flat, @{bvn}, true)"

    def not_term(cursor: str, t: NotTerm, context_path: Optional[str] = None) -> Tuple[Optional[str], str]:
        pre, term_string = term(cursor, t.term, context_path)
        return pre, f"NOT ({term_string})"

    def term(cursor: str, ab_term: Term, context_path: Optional[str] = None) -> Tuple[Optional[str], str]:
        if isinstance(ab_term, AllTerm):
            return None, "true"
        if isinstance(ab_term, Predicate):
            return predicate(cursor, ab_term, context_path)
        if isinstance(ab_term, ContextTerm):
            return context_term(cursor, ab_term, context_path)
        elif isinstance(ab_term, FunctionTerm):
            return None, as_arangodb_function(cursor, bind_vars, ab_term, query_model)
        elif isinstance(ab_term, IdTerm):
            return None, with_id(cursor, ab_term)
        elif isinstance(ab_term, IsTerm):
            return None, is_term(cursor, ab_term)
        elif isinstance(ab_term, NotTerm):
            return not_term(cursor, ab_term, context_path)
        elif isinstance(ab_term, FulltextTerm):
            return None, fulltext_term(cursor, ab_term)
        elif isinstance(ab_term, CombinedTerm):
            pre_left, left = term(cursor, ab_term.left, context_path)
            pre_right, right = term(cursor, ab_term.right, context_path)
            pre = pre_left + " " + pre_right if pre_left and pre_right else pre_left if pre_left else pre_right
            return pre, f"({left}) {ab_term.op} ({right})"
        else:
            raise AttributeError(f"Do not understand: {ab_term}")

    def merge(cursor: str, merge_queries: List[MergeQuery]) -> Tuple[str, str]:  # cursor, query
        result_cursor = next_crs("merge_result")
        merge_cursor = next_crs()
        merge_result = f"LET {result_cursor} = (FOR {merge_cursor} in {cursor} "
        merge_parts: Json = {}

        def add_merge_query(mq: MergeQuery, part_result: str) -> None:
            nonlocal merge_result
            # make sure the sub query is valid
            f = mq.query.parts[-1]
            assert (
                f.term == AllTerm() and not f.sort and not f.limit and not f.with_clause and not f.tag
            ), "Merge query needs to start with navigation!"
            merge_crsr = next_crs("merge_part")
            # make sure the limit only yields one element
            mg_crs, mg_query = query_string(
                db, mq.query, query_model, merge_cursor, with_edges, bind_vars, counters, merge_crsr
            )
            if mq.only_first:
                merge_result += (
                    f"LET {part_result}=FIRST({mg_query} FOR r in {mg_crs} LIMIT 1 RETURN UNSET(r, {unset_props}))"
                )
            else:
                merge_result += (
                    f"LET {part_result}=({mg_query} FOR r in {mg_crs} RETURN DISTINCT UNSET(r, {unset_props}))"
                )

        # check if this query points to an already resolved value
        # Currently only resolved ancestors are taken into account:
        # <-[1:]- is(cloud|account|region|zone)
        # noinspection PyUnresolvedReferences
        def is_already_resolved(q: Query) -> Optional[str]:
            def check_is(t: IsTerm) -> Optional[str]:
                for kind in t.kinds:
                    if kind in GraphResolver.resolved_ancestors:
                        return kind
                return None

            # noinspection PyTypeChecker
            return (
                check_is(q.parts[0].term)
                if (
                    len(q.parts) == 2
                    and not q.aggregate
                    and q.parts[1].navigation
                    and q.parts[1].navigation.direction == "in"
                    and q.parts[1].navigation.until > 1
                    and isinstance(q.parts[0].term, IsTerm)
                )
                else None
            )

        for mq_in in merge_queries:
            part_res = next_crs("part_res")
            resolved = is_already_resolved(mq_in.query)
            if resolved:
                merge_result += f'LET {part_res} = DOCUMENT("{db.vertex_name}", {merge_cursor}.refs.{resolved}_id)'
            else:
                add_merge_query(mq_in, part_res)
            set_value_in_path(part_res, mq_in.name, merge_parts)

        def merge_part_result(d: Json) -> str:
            vals = [f"{k}: {merge_part_result(v)}" if isinstance(v, dict) else f"{k}: {v}" for k, v in d.items()]
            return "{" + ", ".join(vals) + "}"

        final_merge = f"RETURN MERGE_RECURSIVE({merge_cursor}, {merge_part_result(merge_parts)}))"
        return result_cursor, f"{merge_result} {final_merge}"

    def part(p: Part, in_cursor: str, part_idx: int) -> Tuple[Part, str, str, str]:
        query_part = ""
        filtered_out = ""

        def filter_statement(current_cursor: str, part_term: Term, limit: Optional[Limit]) -> str:
            if isinstance(part_term, AllTerm) and limit is None and not p.sort:
                return current_cursor
            nonlocal query_part, filtered_out
            crsr = next_crs()
            filtered_out = next_crs("filter")
            md = f"NOT_NULL({crsr}.metadata, {{}})"
            limited = f" LIMIT {limit.offset}, {limit.length} " if limit else " "
            pre, term_string = term(crsr, part_term)
            pre_string = pre + " FILTER" if pre else "FILTER"
            for_stmt = f"FOR {crsr} in {current_cursor} {pre_string} {term_string}"
            # in case nested properties get unfolded, we need to make the list distinct again
            if pre:
                nested_distinct = next_crs("nested_distinct")
                for_stmt = f"LET {nested_distinct} = ({for_stmt} RETURN DISTINCT {crsr})"
                crsr = next_crs()
                sort_by = sort(crsr, p.sort) if p.sort else " "
                for_stmt = f"{for_stmt} FOR {crsr} in {nested_distinct}{sort_by}{limited}"
            else:
                sort_by = sort(crsr, p.sort) if p.sort else " "
                for_stmt = f"{for_stmt}{sort_by}{limited}"
            f_res = f'MERGE({crsr}, {{metadata:MERGE({md}, {{"query_tag": "{p.tag}"}})}})' if p.tag else crsr
            return_stmt = f"RETURN {f_res}"
            reverse = "REVERSE" if p.reverse_result else ""
            query_part += f"LET {filtered_out} = {reverse}({for_stmt}{return_stmt})"
            return filtered_out

        def with_usage(in_crsr: str, usage: WithUsage, term: Term, limit: Optional[Limit]) -> str:
            nonlocal query_part

            # split the term and create a filter statement for everything before the usage predicates
            before_term, after_term = term.split_by_usage()

            # the limit is applied here, when the after term does not filter at all
            after_filter_cursor = filter_statement(in_crsr, before_term, limit=limit if after_term.is_all else None)

            # add the usage predicates
            usage_crs = next_crs("with_usage")
            start = next_bind_var_name()
            end = next_bind_var_name()
            start_s = next_bind_var_name()
            duration = next_bind_var_name()
            start_time = usage.start_from_now()
            end_time = usage.end_from_now()
            bind_vars[start] = start_time.timestamp()
            bind_vars[end] = end_time.timestamp()
            bind_vars[start_s] = utc_str(start_time)
            bind_vars[duration] = duration_str(end_time - start_time)
            avgs = []
            merges = []
            for mn in usage.metrics:
                avgs.append(f"{mn}_min = MIN(m.v.{mn}.min), {mn}_avg = AVG(m.v.{mn}.avg), {mn}_max = MAX(m.v.{mn}.max)")
                merges.append(f"{mn}: {{min: {mn}_min, avg: {mn}_avg, max: {mn}_max}}")
            query_part += dedent(
                f"""
                let {usage_crs} = (
                    for r in {after_filter_cursor}
                        let resource=r
                        let resource_usage = first(
                            for m in {db.usage_db.collection_name}
                            filter m.at>=@{start} and m.at<=@{end} and m.id==r._key
                            collect aggregate {", ".join(avgs)}, count = sum(1)
                            return {{usage:{{{",".join(merges)},entries:count,start:@{start_s},duration:@{duration}}}}}
                        )
                        return resource_usage.usage.entries ? merge(resource, resource_usage) : resource
                )
                """
            )

            # finally apply the filter that includes the usage predicates
            return filter_statement(usage_crs, after_term, limit)

        def with_clause(in_crsr: str, clause: WithClause, limit: Optional[Limit]) -> str:
            nonlocal query_part
            # this is the general structure of the with_clause that is created
            #
            # FOR cloud in foo FILTER @0 in cloud.kinds
            #    FOR account IN 0..1 OUTBOUND cloud foo_default
            #    OPTIONS { bfs: true, uniqueVertices: 'global' }
            #    FILTER (cloud._key==account._key) or (@1 in account.kinds)
            #        FOR region in 0..1 OUTBOUND account foo_default
            #        OPTIONS { bfs: true, uniqueVertices: 'global' }
            #         FILTER (cloud._key==region._key) or (@2 in region.kinds)
            #             FOR zone in 0..1 OUTBOUND region foo_default
            #             OPTIONS { bfs: true, uniqueVertices: 'global' }
            #             FILTER (cloud._key==zone._key) or (@3 in zone.kinds)
            #         COLLECT l4_cloud = cloud, l4_account=account, l4_region=region WITH COUNT INTO counter3
            #         FILTER (l4_cloud._key==l4_region._key) or (counter3>=0)
            #     COLLECT l3_cloud = l4_cloud, l3_account=l4_account WITH COUNT INTO counter2
            #     FILTER (l3_cloud._key==l3_account._key) or (counter2>=0) // ==2 regions
            # COLLECT l2_cloud = l3_cloud WITH COUNT INTO counter1
            # FILTER (counter1>=0) //counter is +1 since the node itself is always bypassed
            # RETURN ({cloud: l2_cloud._key, count:counter1})
            current = next_counter("with_clause")

            def cursor_in(depth: int) -> str:
                return f"c{current}_{depth}"

            l0crsr = cursor_in(0)

            def traversal_filter(cl: WithClause, in_crs: str, depth: int) -> str:
                nav = cl.navigation
                crsr = cursor_in(depth)
                direction = "OUTBOUND" if nav.direction == Direction.outbound else "INBOUND"
                unique = "uniqueEdges: 'path'" if with_edges else "uniqueVertices: 'global'"
                pre, term_string = term(crsr, cl.term) if cl.term else (None, "true")
                pre_string = pre + " FILTER" if pre else "FILTER"
                filter_clause = f"({term_string})"
                inner = traversal_filter(cl.with_clause, crsr, depth + 1) if cl.with_clause else ""
                filter_root = f"({l0crsr}._key=={crsr}._key) or " if depth > 0 else ""
                edge_type_traversals = f", {direction} ".join(f"`{db.edge_collection(et)}`" for et in nav.edge_types)
                return (
                    f"FOR {crsr} IN 0..{nav.until} {direction} {in_crs} "
                    f"{edge_type_traversals} OPTIONS {{ bfs: true, {unique} }} "
                    f"{pre_string} {filter_root}{filter_clause} "
                ) + inner

            def collect_filter(cl: WithClause, depth: int) -> str:
                fltr = cl.with_filter
                if cl.with_clause:
                    collects = ", ".join(f"l{depth-1}_l{i}_res=l{depth}_l{i}_res" for i in range(0, depth))
                else:
                    collects = ", ".join(f"l{depth-1}_l{i}_res={cursor_in(i)}" for i in range(0, depth))

                if depth == 1:
                    # note: the traversal starts from 0 (only 0 and 1 is allowed)
                    # when we start from 1: increase the count by one to not count the start node
                    # when we start from 0: the start node is expected in the count already
                    filter_term = f"FILTER counter1{fltr.op}{fltr.num + cl.navigation.start}"
                else:
                    root_key = f"l{depth-1}_l0_res._key==l{depth-1}_l{depth-1}_res._key"
                    filter_term = f"FILTER ({root_key}) or (counter{depth}{fltr.op}{fltr.num})"

                inner = collect_filter(cl.with_clause, depth + 1) if cl.with_clause else ""
                return inner + f"COLLECT {collects} WITH COUNT INTO counter{depth} {filter_term} "

            out = next_crs()

            limited = f" LIMIT {limit.offset}, {limit.length} " if limit else " "
            query_part += (
                f"LET {out} =( FOR {l0crsr} in {in_crsr} "
                + traversal_filter(clause, l0crsr, 1)
                + collect_filter(clause, 1)
                + limited
                + "RETURN l0_l0_res) "
            )
            return out

        def inout(in_crsr: str, start: int, until: int, edge_type: str, direction: str) -> str:
            nonlocal query_part
            in_c = next_crs("io_in")
            out = next_crs("io_out")
            out_crsr = next_crs("io_crs")
            link = next_crs("io_link")
            unique = "uniqueEdges: 'path'" if with_edges else "uniqueVertices: 'global'"
            link_str = f", {link}" if with_edges else ""
            dir_bound = "OUTBOUND" if direction == Direction.outbound else "INBOUND"
            inout_result = (
                f"MERGE({out_crsr}, {{_from:{link}._from, _to:{link}._to, _link_id:{link}._id}})"
                if with_edges
                else out_crsr
            )
            if outer_merge and part_idx == 0:
                graph_cursor = in_crsr
                outer_for = ""
            else:
                graph_cursor = in_c
                outer_for = f"FOR {in_c} in {in_crsr} "

            query_part += (
                f"LET {out} =({outer_for}"
                f"FOR {out_crsr}{link_str} IN {start}..{until} {dir_bound} {graph_cursor} "
                f"`{db.edge_collection(edge_type)}` OPTIONS {{ bfs: true, {unique} }} "
                f"RETURN DISTINCT {inout_result}) "
            )
            return out

        def navigation(in_crsr: str, nav: Navigation) -> str:
            nonlocal query_part
            all_walks = []
            if nav.direction == Direction.any:
                for et in nav.edge_types:
                    all_walks.append(inout(in_crsr, nav.start, nav.until, et, Direction.inbound))
                for et in nav.maybe_two_directional_outbound_edge_type or nav.edge_types:
                    all_walks.append(inout(in_crsr, nav.start, nav.until, et, Direction.outbound))
            else:
                for et in nav.edge_types:
                    all_walks.append(inout(in_crsr, nav.start, nav.until, et, nav.direction))

            if len(all_walks) == 1:
                return all_walks[0]
            else:
                nav_crsr = next_crs()
                all_walks_combined = ",".join(all_walks)
                query_part += f"LET {nav_crsr} = UNION_DISTINCT({all_walks_combined})"
                return nav_crsr

        # apply the limit in the filter statement only, when no with clause is present
        # otherwise the limit is applied in the with clause
        filter_limit = p.limit if p.with_clause is None else None
        cursor = in_cursor
        part_term = p.term
        if isinstance(p.term, MergeTerm):
            # only allow a limit in the prefilter, if there is no post filter
            pre_limit = filter_limit if (p.term.post_filter is None or p.term.post_filter.is_all) else None
            filter_cursor = filter_statement(cursor, p.term.pre_filter, pre_limit)
            cursor, merge_part = merge(filter_cursor, p.term.merge)
            query_part += merge_part
            # always do the post filter in case of sort or limit
            part_term = p.term.post_filter if p.term.post_filter else AllTerm()
        if p.with_usage and len(p.with_usage.metrics) > 0:
            # filter is applied in the with usage
            cursor = with_usage(cursor, p.with_usage, part_term, filter_limit)
        else:
            cursor = filter_statement(cursor, part_term, filter_limit)

        cursor = with_clause(cursor, p.with_clause, p.limit) if p.with_clause else cursor
        cursor = navigation(cursor, p.navigation) if p.navigation else cursor
        return p, cursor, filtered_out, query_part

    def sort(cursor: str, so: List[Sort]) -> str:
        def single_sort(single: Sort) -> str:
            prop_name, resolved, _ = prop_name_kind(single.name)
            order = SortOrder.reverse(single.order) if resolved.simple_kind.reverse_order else single.order
            return f"{cursor}.{prop_name} {order}"

        sorts = ", ".join(single_sort(s) for s in so)
        return f" SORT {sorts} "

    def fulltext(ft_part: Term, filter_term: Term) -> Tuple[str, str]:
        # The fulltext index only understands not, combine and fulltext
        def ft_term(cursor: str, ab_term: Term) -> str:
            if isinstance(ab_term, NotTerm):
                return f"NOT ({ft_term(cursor, ab_term.term)})"
            elif isinstance(ab_term, FulltextTerm):
                bvn = next_bind_var_name()
                bind_vars[bvn] = ab_term.text
                # the fulltext index is based on the flat property. The full text term is tokenized.
                return f"PHRASE({cursor}.flat, @{bvn})"
            elif isinstance(ab_term, CombinedTerm):
                left = ft_term(cursor, ab_term.left)
                right = ft_term(cursor, ab_term.right)
                return f"({left}) {ab_term.op} ({right})"
            else:
                raise AttributeError(f"Do not understand: {ab_term}")

        # Since fulltext filtering is handled separately, we replace the remaining filter term in the first part
        query_parts[0] = evolve(query_parts[0], term=filter_term)
        crs = next_crs()
        doc = f"search_{db.vertex_name}"
        ftt = ft_term("ft", ft_part)
        q = f"LET {crs}=(FOR ft in {doc} SEARCH ANALYZER({ftt}, 'delimited') SORT BM25(ft) DESC RETURN ft)"
        return q, crs

    parts = []
    ft, remaining = fulltext_term_combine(query_parts[0].term)
    fulltext_part, crsr = fulltext(ft, remaining) if ft else ("", start_cursor)
    for idx, p in enumerate(query_parts):
        part_tuple = part(p, crsr, idx)
        parts.append(part_tuple)
        crsr = part_tuple[1]

    query_str = fulltext_part + " ".join(p[3] for p in parts)
    resulting_cursor = crsr
    nxt = next_crs()
    if query.aggregate:  # return aggregate
        resulting_cursor, aggregation = aggregate(resulting_cursor, query.aggregate)
        query_str += aggregation
        # if the last part has a sort order, we use it here again
        if query.current_part.sort:
            sort_by = sort("res", query.current_part.sort)
            query_str += f" LET {nxt} = (FOR res in {resulting_cursor}{sort_by} RETURN res)"
            resulting_cursor = nxt
    else:  # return results
        # return all tagged commands (last result is "tagged" automatically)
        tagged = {out for part, _, out, _ in parts if part.tag}
        if tagged:
            tagged_union = f'UNION({",".join(tagged)},{resulting_cursor})'
            query_str += f" LET {nxt} = (FOR res in {tagged_union} RETURN res)"
            resulting_cursor = nxt
    return resulting_cursor, query_str


async def query_cost(graph_db: Any, model: QueryModel, with_edges: bool) -> EstimatedSearchCost:
    q_string, bind = to_query(graph_db, model, with_edges=with_edges)
    nr_nodes = await graph_db.db.count(graph_db.vertex_name)
    plan = await graph_db.db.explain(query=q_string, bind_vars=bind)
    full_collection_scan = exist(lambda node: node["type"] == "EnumerateCollectionNode", plan["nodes"])
    estimated_cost = int(plan["estimatedCost"])
    estimated_items = int(plan["estimatedNrItems"])
    # If the number of returned items is small, most of the computation happens on the db side
    # A higher factor (==estimated cost) is acceptable in this case.
    factor = 20 if estimated_items < 3 else 2.1
    # max upper bound, if the number of nodes is very small
    ratio = estimated_cost / max(10000, nr_nodes * factor)
    # the best rating is complex, if a full collection scan is required.
    best = Rating.complex if full_collection_scan else Rating.simple
    rating = best if ratio < 0.2 else (Rating.complex if ratio < 1 else Rating.bad)
    return EstimatedSearchCost(estimated_cost, estimated_items, nr_nodes, full_collection_scan, rating)


def fulltext_term_combine(term_in: Term) -> Tuple[Optional[Term], Term]:
    """
    Split the term of this part into the independent fulltext term and the remaining part of the term.
    Logic: self.term ~=logical_equivalent=~ fulltext & remaining
    :return: a term that can utilize the fulltext search index and a "normal" filter term.
    """

    def combine_fulltext(term: Term) -> Tuple[Term, Term]:
        if not term.contains_term_type(FulltextTerm):
            return AllTerm(), term
        elif isinstance(term, FulltextTerm):
            return term, AllTerm()
        elif isinstance(term, CombinedTerm):
            if (
                (term.left.contains_term_type(FulltextTerm) or term.right.contains_term_type(FulltextTerm))
                and term.op == "or"
                and term.find_term(lambda x: not isinstance(x, FulltextTerm) and not isinstance(x, CombinedTerm))
            ):
                # This term can not utilize the search index!
                return AllTerm(), term
            left = isinstance(term.left, FulltextTerm)
            right = isinstance(term.right, FulltextTerm)
            if left and right:
                return term, AllTerm()
            elif left:
                ft, remaining = combine_fulltext(term.right)
                return ft.combine(term.op, term.left), remaining
            elif right:
                ft, remaining = combine_fulltext(term.left)
                return ft.combine(term.op, term.right), remaining
            else:
                lf, remaining_left = combine_fulltext(term.right)
                rf, remaining_right = combine_fulltext(term.left)
                return lf.combine(term.op, rf), remaining_left.combine(term.op, remaining_right)
        elif isinstance(term, NotTerm):
            ft, remaining = combine_fulltext(term.term)
            return NotTerm(ft), remaining if isinstance(remaining, AllTerm) else NotTerm(remaining)
        elif isinstance(term, MergeTerm):
            ft, remaining = combine_fulltext(term.pre_filter)
            return ft, evolve(term, pre_filter=remaining)
        else:
            raise AttributeError(f"Can not handle term of type: {type(term)} ({term})")

    fulltext, new_term = combine_fulltext(term_in)
    return (None, term_in) if isinstance(fulltext, AllTerm) else (fulltext, new_term)



from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, EventEntityDb, ArangoEntityDb
from resotocore.query.model import Template

TemplateEntityDb = EntityDb[str, Template]
EventTemplateEntityDb = EventEntityDb[str, Template]


def template_entity_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[str, Template]:
    return ArangoEntityDb(db, collection, Template, lambda k: k.name)



from typing import Optional

from networkx import MultiDiGraph

from tests.resotocore.db.graphdb_test import Foo, Bla
from resotocore.model.graph_access import EdgeTypes, GraphAccess
from resotocore.model.typed_model import to_js
from resotocore.types import EdgeType, Json, JsonElement


def create_graph(bla_text: str, width: int = 10) -> MultiDiGraph:
    graph = MultiDiGraph()

    def add_edge(from_node: str, to_node: str, edge_type: EdgeType = EdgeTypes.default) -> None:
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        graph.add_edge(from_node, to_node, key, edge_type=edge_type)

    def add_node(uid: str, kind: str, node: Optional[Json] = None, replace: bool = False) -> None:
        reported = {**(node if node else to_js(Foo(uid))), "kind": kind}
        graph.add_node(
            uid,
            id=uid,
            kinds=[kind],
            reported=reported,
            desired={"node_id": uid},
            metadata={"node_id": uid},
            replace=replace,
            ancestors={
                "cloud": {"reported": {"name": "collector", "id": "collector"}},
                "account": {"reported": {"name": "sub_root", "id": "sub_root"}},
            },
            refs={
                "cloud_id": "collector",
                "account_id": "sub_root",
            },
        )

    # root -> collector -> sub_root -> **rest
    add_node("root", "graph_root")
    add_node("collector", "cloud", replace=True)
    add_node("sub_root", "foo")
    add_edge("root", "collector")
    add_edge("collector", "sub_root")

    for o in range(0, width):
        oid = str(o)
        add_node(oid, "foo")
        add_edge("sub_root", oid)
        for i in range(0, width):
            iid = f"{o}_{i}"
            add_node(iid, "bla", node=to_js(Bla(iid, name=bla_text)))
            add_edge(oid, iid)
            add_edge(iid, oid, EdgeTypes.delete)
    return graph



from typing import Tuple, Dict, Optional

from resotocore.db import SystemData, drop_arango_props
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.model.typed_model import from_js
from resotocore.util import if_set


class SystemDataDb:
    def __init__(self, db: AsyncArangoDB):
        self.db = db
        self.collection_name = "system_data"

    async def system_data(self) -> SystemData:
        return if_set(await self.db.get(self.collection_name, "system"), lambda x: from_js(x, SystemData))  # type: ignore # noqa E501

    async def ca(self) -> Tuple[str, str]:
        return if_set(await self.db.get(self.collection_name, "ca"), lambda x: (x["key"], x["certificate"]))  # type: ignore # noqa E501

    async def jwt_signing_keys(self) -> Optional[Tuple[str, str]]:
        return if_set(await self.db.get(self.collection_name, "jwt_signing_keys"), lambda x: (x["key"], x["certificate"]))  # type: ignore # noqa E501

    async def info(self) -> Dict[str, str]:
        return drop_arango_props((await self.db.get(self.collection_name, "info")) or {})

    async def update_jwt_signing_keys(self, key: str, certificate: str) -> Tuple[str, str]:
        doc = await self.db.insert(
            self.collection_name,
            {"_key": "jwt_signing_keys", "key": key, "certificate": certificate},
            return_new=True,
            overwrite=True,
            overwrite_mode="replace",
        )
        return doc["new"]["key"], doc["new"]["certificate"]  # type: ignore

    async def update_info(self, **kwargs: str) -> Dict[str, str]:
        kwargs["_key"] = "info"
        doc = await self.db.insert(
            self.collection_name, kwargs, return_new=True, overwrite=True, overwrite_mode="update", merge=True
        )
        return drop_arango_props(doc["new"])  # type: ignore
